import numpy as np
from datetime import datetime
from pathlib import Path
from pdb import set_trace
import sys
import os
from os.path import join
from scipy.io import loadmat
# To test logger:
import cvnn
import logging

logger = logging.getLogger(cvnn.__name__)


def load_matlab_matrices(fname="data_cnn1dT.mat", path="/media/barrachina/data/gilles_data/"):
    """
    Opens Matlab matrix (.mat) as numpy array.
    :param fname: file name to be opened
    :param path: path to file
    :return: numpy array with the Matlab matrix information
    """
    mat_fname = join(path, fname)
    mat = loadmat(mat_fname)
    return mat


def create_folder(root_path, now=None):
    """
    Creates folders within root_path using a date format.
    :param root_path: root path where to create the folder chain
    :param now: date to be used. If None then it will use current time
    :return: the created path in pathlib format (compatible across different OS)
    """
    if now is None:
        now = datetime.today()
    # path = Path(__file__).parents[1].absolute() / Path(root_path + now.strftime("%Y/%m%B/%d%A/run-%Hh%Mm%S/"))
    # Last line was to create inside cvnn. I prefer now to save stuff on each project folder and not on libraries folder
    path = Path(root_path + now.strftime("%Y/%m%B/%d%A/run-%Hh%Mm%S/"))
    os.makedirs(path, exist_ok=True)        # Do this not to have a problem if I run in parallel
    return path


def cast_to_path(path):
    if isinstance(path, str):
        path = Path(path)
    elif not isinstance(path, Path):
        logger.error("Path datatype not recognized")
        sys.exit(-1)
    return path


def get_func_name(fun):
    """
    Returns the name of a function passed as parameter being either a function itself or a string with the function name
    :param fun: function or function name
    :return: function name
    """
    if callable(fun):
        return fun.__name__
    elif isinstance(fun, str):
        return fun
    else:
        logger.error("Function not recognizable", stack_info=True)
        sys.exit(-1)


def transform_to_real(x_complex, polar=False):
    """
    Transforms a complex input matrix into a real value matrix (double size)
    :param x_complex: Complex-valued matrix of size mxn
    :param polar: If True, the data returned will be the amplitude and phase instead of real an imaginary part
        (Default: False)
    :return: real-valued matrix of size mx(2*n) unwrapping the real and imag part of the complex-valued input matrix
    """
    # import pdb; pdb.set_trace()
    m = np.shape(x_complex)[0]
    n = np.shape(x_complex)[1]
    x_real = np.ones((m, 2*n))
    if not polar:
        x_real[:, :n] = np.real(x_complex)
        x_real[:, n:] = np.imag(x_complex)
    else:
        x_real[:, :n] = np.abs(x_complex)
        x_real[:, n:] = np.angle(x_complex)
    dtype = x_real.dtype
    if x_complex.dtype == np.complex64:
        dtype = np.float32
    elif x_complex.dtype == np.complex128:
        dtype = np.float64
    else:
        logger.warning("data type unknown: " + str(x_complex.dtype))
    return x_real.astype(dtype)


def cart2polar(z):
    """
    :param z: complex input
    :return: tuple with the absolute value of the input and the phase
    """
    return np.abs(z), np.angle(z)


def polar2cart(rho, angle):
    """
    :param rho: absolute value
    :param angle: phase
    :return: complex number using phase and angle
    """
    return rho * np.exp(1j*angle)


def randomize(x, y):
    """
    Randomizes the order of data samples and their corresponding labels
    :param x: data
    :param y: data labels
    :return: Tuple of (shuffled_x, shuffled_y) maintaining coherence of elements labels
    """
    permutation = np.random.permutation(y.shape[0])
    shuffled_x = x[permutation, :]
    shuffled_y = y[permutation]
    return shuffled_x, shuffled_y


def normalize(x):
    return (x-np.amin(x))/np.abs(np.amax(x)-np.amin(x))     # Checked it works for complex values


def tensorflow_argmax_np_equivalent(x, num_classes):
    res = np.zeros((np.argmax(x, 1).shape[0], num_classes))
    indx = 0
    for k in np.argmax(x, 1):
        res[indx, k] = 1
        indx += 1
    return res


def compute_accuracy(x, y):
    return np.average(np.equal(x, y).all(axis=1))


if __name__ == "__main__":
    logger.warning("Testing logger")


__author__ = 'J. Agustin BARRACHINA'
__version__ = '0.0.14'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'



from dataclasses import dataclass
import matplotlib.pyplot as plt
import plotly
import plotly.graph_objects as go
import plotly.figure_factory as ff
import seaborn as sns
import pandas as pd
import numpy as np
import re
import os
import sys
from pathlib import Path
from pdb import set_trace
from cvnn.utils import create_folder
import logging
import cvnn
import tikzplotlib

logger = logging.getLogger(cvnn.__name__)

DEFAULT_PLOTLY_COLORS = ['rgb(31, 119, 180)',  # Blue
                         'rgb(255, 127, 14)',  # Orange
                         'rgb(44, 160, 44)',  # Green
                         'rgb(214, 39, 40)',
                         'rgb(148, 103, 189)', 'rgb(140, 86, 75)',
                         'rgb(227, 119, 194)', 'rgb(127, 127, 127)',
                         'rgb(188, 189, 34)', 'rgb(23, 190, 207)']

DEFAULT_MATPLOTLIB_COLORS = plt.rcParams['axes.prop_cycle'].by_key()['color']  # [1:] # Uncomment to remove blue color


@dataclass
class Resolution:
    width: int
    height: int


RESOLUTIONS_16_9 = {
    'lowest': Resolution(1024, 576),
    'low': Resolution(1152, 648),
    'HD': Resolution(1280, 720),  # 720p
    'FHD': Resolution(1920, 1080),  # 1080p
    'QHD': Resolution(2560, 1440),  # 1440p
    'UHD': Resolution(2560, 1440)  # 4K or 2160p
}
RESOLUTIONS_4_3 = {
    '640×480': Resolution(640, 480),
    '800×600': Resolution(800, 600),
    '960×720': Resolution(960, 720),
    '1024×768': Resolution(1024, 768),
    '1280×960': Resolution(1280, 960),
    # https://www.comtech-networking.com/blog/item/4-what-is-the-screen-resolution-or-the-aspect-ratio-what-do-720p-1080i-1080p-mean/
}

PLOTLY_CONFIG = {
    'scrollZoom': True,
    'editable': True,
    'toImageButtonOptions': {
        'format': 'svg',  # one of png, svg, jpeg, webp
        # 'filename': 'custom_image',
        'height': RESOLUTIONS_4_3['800×600'].height,
        'width': RESOLUTIONS_4_3['800×600'].width,
        'scale': 1  # Multiply title/legend/axis/canvas sizes by this factor
    }
}


def triangulate_histogram(x, y, z):
    # https://community.plot.ly/t/adding-a-shape-to-a-3d-plot/1441/8?u=negu93
    if len(x) != len(y) != len(z):
        raise ValueError("The  lists x, y, z, must have the same length")
    n = len(x)
    if n % 2:
        raise ValueError("The length of lists x, y, z must be an even number")
    pts3d = np.vstack((x, y, z)).T
    pts3dp = np.array([[x[2 * k + 1], y[2 * k + 1], 0] for k in range(1, n // 2 - 1)])
    pts3d = np.vstack((pts3d, pts3dp))
    # triangulate the histogram bars:
    tri = [[0, 1, 2], [0, 2, n]]
    for k, i in zip(list(range(n, n - 3 + n // 2)), list(range(3, n - 4, 2))):
        tri.extend([[k, i, i + 1], [k, i + 1, k + 1]])
    tri.extend([[n - 3 + n // 2, n - 3, n - 2], [n - 3 + n // 2, n - 2, n - 1]])
    return pts3d, np.array(tri)


def add_transparency(color='rgb(31, 119, 180)', alpha=0.5):
    pattern = re.compile("^rgb\([0-9]+, [0-9]+, [0-9]+\)$")
    if not re.match(pattern, color):
        logger.error("Unrecognized color format")
        sys.exit(-1)
    color = re.sub("^rgb", "rgba", color)
    color = re.sub("\)$", ", {})".format(alpha), color)
    return color


def extract_values(color='rgb(31, 119, 180)'):
    pattern = re.compile("^rgb\([0-9]+, [0-9]+, [0-9]+\)$")
    if not re.match(pattern, color):
        logger.error("Unrecognized color format")
        sys.exit(-1)
    return [float(s) for s in re.findall(r'\b\d+\b', color)]


def find_intersection_of_gaussians(m1, m2, std1, std2):
    a = 1 / (2 * std1 ** 2) - 1 / (2 * std2 ** 2)
    b = m2 / (std2 ** 2) - m1 / (std1 ** 2)
    c = m1 ** 2 / (2 * std1 ** 2) - m2 ** 2 / (2 * std2 ** 2) - np.log(std2 / std1)
    return np.roots([a, b, c])


def add_params(fig, ax, y_label=None, x_label=None, loc=None, title=None,
               filename="./results/plot_2_gaussian_output.png", showfig=False, savefig=True):
    """
    :param fig:
    :param ax:
    :param y_label: The y axis label.
    :param x_label: The x axis label.
    :param loc: can be a string or an integer specifying the legend location. default: None.
                    https://matplotlib.org/api/legend_api.html#matplotlib.legend.Legend
    :param title: str or None. The legend’s title. Default is no title (None).
    :param filename: Only used when savefig=True. The name of the figure to be saved
    :param showfig: Boolean. If true it will show the figure using matplotlib show method
    :param savefig: Boolean. If true it will save the figure with the name of filename parameter
    :return None:
    """
    # Figure parameters
    if loc is not None:
        fig.legend(loc=loc)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if x_label is not None:
        ax.set_xlabel(x_label)
    if title is not None:
        ax.set_title(title)
    # save/show results
    if showfig:
        fig.show()
    if savefig:
        os.makedirs(os.path.split(filename)[0], exist_ok=True)
        fig.savefig(filename, transparent=True)


def get_trailing_number(s):
    """
    Search for a termination of a file name that has the ".csv" extension and that has a number at the end.
    It gives the number at the end of the file. This number can have any amount of digits.
    Example:
    x = get_trailing_number("my/path/to/file/any_43_start_name9872.csv")    # x = 9872
    y = get_trailing_number("my/path/to/file/any_43_start_name.csv")        # y = None
    z = get_trailing_number("my/path/to/file/any_43_start_name85498.txt")   # y = None
    :param s: The string to search for the specific term
    :return: The number located before the extension. None if there is no number.
    """
    m = re.search(r'\d+.csv$', s)  # I get only the end of the string (last number of any size and .csv extension)
    # splitext gets root [0] and extension [1] of the name.
    return int(os.path.splitext(m.group())[0]) if m else None


# ----------------
# Confusion Matrix
# ----------------


def plot_confusion_matrix(data, filename=None, library='plotly', axis_legends=None, showfig=False):
    if library == 'seaborn':
        fig, ax = plt.subplots()
        sns.heatmap(data,
                    annot=True,
                    linewidths=.5,
                    cbar=True,
                    )
        if filename is not None:
            fig.savefig(filename)
    elif library == 'plotly':
        z = data.values.tolist()
        if axis_legends is None:
            y = [str(j) for j in data.axes[0].tolist()]
            x = [str(i) for i in data.axes[1].tolist()]
        else:
            y = []
            x = []
            for j in data.axes[0].tolist():
                if isinstance(j, int):
                    y.append(axis_legends[j])
                elif isinstance(j, str):
                    y.append(j)
                else:
                    logger.critical("WTF?! should never have arrived here")
            for i in data.axes[1].tolist():
                if isinstance(i, int):
                    x.append(axis_legends[i])
                elif isinstance(i, str):
                    x.append(i)
                else:
                    logger.critical("WTF?! should never have arrived here")
        # fig = go.Figure(data=go.Heatmap(z=z, x=x, y=y))
        fig = ff.create_annotated_heatmap(z, x=x, y=y)
    if showfig:
        fig.show()


def confusion_matrix(y_pred_np, y_label_np, filename=None, axis_legends=None):
    categorical = (len(np.shape(y_label_np)) > 1)
    if categorical:
        y_pred_np = np.argmax(y_pred_np, axis=1)
        y_label_np = np.argmax(y_label_np, axis=1)
    y_pred_pd = pd.Series(y_pred_np, name='Predicted')
    y_label_pd = pd.Series(y_label_np, name='Actual')
    df = pd.crosstab(y_label_pd, y_pred_pd, rownames=['Actual'], colnames=['Predicted'], margins=True)
    if filename is not None:
        df.to_csv(filename)
    # plot_confusion_matrix(df, filename, library='plotly', axis_legends=axis_legends)
    return df


# ----------------
# Comparison
# ----------------


class SeveralMonteCarloComparison:

    def __init__(self, label, x, paths, round=2):
        """
        This class is used to compare several monte carlo runs done with cvnn.montecarlo.MonteCarlo class.
        MonteCarlo let's you compare different models between them but let's you not change other values like epochs.
        You can run as several MonteCarlo runs and then use SeveralMonteCarloComparison class to compare the results.

        Example of usage:

        ```
        # Run several Monte Carlo's
        for learning_rate in learning_rates:
            monte_carlo = RealVsComplex(complex_network)
            monte_carlo.run(x, y, iterations=iterations, learning_rate=learning_rate,
                            epochs=epochs, batch_size=batch_size, display_freq=display_freq,
                            shuffle=True, debug=debug, data_summary=dataset.summary())
        # Run self
        several = SeveralMonteCarloComparison('learning rate', x = learning_rates,
                                              paths = ["path/to/1st/run/run_data",
                                                       "path/to/2nd/run/run_data",
                                                       "path/to/3rd/run/run_data",
                                                       "path/to/4th/run/run_data"]
        several.box_plot(showfig=True)
        ```

        :label: string that describes what changed between each montecarlo run
        :x: List of the value for each monte carlo run wrt :label:.
        :paths: Full path to each monte carlo run_data saved file (Must end with run_data)
            NOTE: x and paths must be the same size
        """
        self.x_label = label
        if all([item.isdigit() for item in x]):
            self.x = list(map(int, x))
        elif all([item.replace(".", "", 1).isdigit() for item in x]):
            self.x = np.round(list(map(float, x)), round)
        else:
            self.x = x
        self.monte_carlo_runs = []
        for path in paths:
            self.monte_carlo_runs.append(MonteCarloAnalyzer(path=path))
        if not len(self.x) == len(self.monte_carlo_runs):
            logger.error("x ({0}) and paths ({1}) must be the same size".format(len(self.x),
                                                                                len(self.monte_carlo_runs)))
        frames = [self.monte_carlo_runs[0].df]
        frames[0]['network'] = frames[0]['network'] + " " + x[0]
        for i, monte_carlo_run in enumerate(self.monte_carlo_runs[1:]):
            frames.append(self.monte_carlo_runs[i + 1].df)
            frames[i + 1]['network'] = frames[i + 1]['network'] + " " + x[i + 1]
        self.df = pd.concat(frames)
        self.monte_carlo_analyzer = MonteCarloAnalyzer(df=self.df)

    def box_plot(self, key='test accuracy', library='plotly', step=-1, showfig=False, savefile=None):
        if library == 'plotly':
            self._box_plot_plotly(key=key, step=step, showfig=showfig, savefile=savefile)
        # TODO: https://seaborn.pydata.org/examples/grouped_boxplot.html
        elif library == 'seaborn':
            self._box_plot_seaborn(key=key, step=step, showfig=showfig, savefile=savefile)
        else:
            logger.warning("Unrecognized library to plot " + library)
        return None

    def _box_plot_plotly(self, key='test accuracy', step=-1, showfig=False, savefile=None):
        # https://en.wikipedia.org/wiki/Box_plot
        # https://plot.ly/python/box-plots/
        # https://towardsdatascience.com/understanding-boxplots-5e2df7bcbd51
        # Median (Q2 / 50th Percentile): Middle value of the dataset. ex. median([1, 3, 3, 6, 7, 8, 9]) = 6
        # First quartile (Q1 / 25th Percentile): Middle value between the median and the min(dataset) = 1
        # Third quartile (Q3 / 75th Percentile): Middle value between the median and the max(dataset) = 9
        # Interquartile Range (IQR) = Q3 - Q1
        # Whishker: [Q1 - 1.5*IQR, Q3 + 1.5*IQR], whatever is out of this is an outlier.
        # suspected outlier: [Q1 - 3*IQR, Q3 + 3*IQR]
        savefig = False
        if savefile is not None:
            savefig = True

        steps = []
        for i in range(len(self.monte_carlo_runs)):
            if step == -1:
                steps.append(max(self.monte_carlo_runs[i].df.step))
            else:
                steps.append(step)

        fig = go.Figure()

        for i, mc_run in enumerate(self.monte_carlo_runs):
            df = mc_run.df
            networks_availables = df.network.unique()
            for color_index, net in enumerate(networks_availables):
                filter = [a == net and b == steps[i] for a, b in zip(df.network, df.step)]
                data = df[filter]
                fig.add_trace(go.Box(
                    y=data[key],
                    # x=[self.x[i]] * len(data[key]),
                    name=net.replace('_', ' ') + " " + str(self.x[i]),
                    whiskerwidth=0.2,
                    notched=True,  # confidence intervals for the median
                    fillcolor=add_transparency(DEFAULT_PLOTLY_COLORS[color_index], 0.5),
                    boxpoints='suspectedoutliers',  # to mark the suspected outliers
                    line=dict(color=DEFAULT_PLOTLY_COLORS[color_index]),
                    boxmean=True  # Interesting how sometimes it falls outside the box
                ))

        fig.update_layout(
            title=self.x_label + ' Box Plot',
            xaxis=dict(title=self.x_label),
            yaxis=dict(
                title=key,
                autorange=True,
                showgrid=True,
                dtick=0.05,
            ),
            # boxmode='group',
            # boxgroupgap=0,
            # boxgap=0,
            showlegend=True
        )
        if savefig:
            if not savefile.endswith('.html'):
                savefile += '.html'
            os.makedirs(os.path.split(savefile)[0], exist_ok=True)
            plotly.offline.plot(fig,
                                filename=savefile, config=PLOTLY_CONFIG, auto_open=showfig)
            # fig.write_image(savefile.replace('.html', extension))
        elif showfig:
            fig.show(config=PLOTLY_CONFIG)

    def _box_plot_seaborn(self, key='test accuracy', step=-1, showfig=False, savefile=None, extension=".svg"):
        steps = []
        for i in range(len(self.monte_carlo_runs)):
            if step == -1:
                steps.append(max(self.monte_carlo_runs[i].df.step))
            else:
                steps.append(step)
        # Prepare data
        frames = []
        for i, mc_run in enumerate(self.monte_carlo_runs):
            df = mc_run.df
            filter = df['step'] == steps[i]
            data = df[filter]
            data[self.x_label] = self.x[i]
            frames.append(data)
        result = pd.concat(frames)

        # Run figure
        fig = plt.figure()
        ax = sns.boxplot(x=self.x_label, y=key, hue="network", data=result, boxprops=dict(alpha=.3))
        # Make black lines the color of the box
        for i, artist in enumerate(ax.artists):
            col = artist.get_facecolor()[:-1]  # the -1 removes the transparency
            artist.set_edgecolor(col)
            for j in range(i * 6, i * 6 + 6):
                line = ax.lines[j]
                line.set_color(col)
                line.set_mfc(col)
                line.set_mec(col)

        if savefile is not None:
            if not savefile.endswith(extension):
                savefile += extension
            os.makedirs(os.path.split(savefile)[0], exist_ok=True)
            fig.savefig(savefile, transparent=True)
            # tikzplotlib.clean_figure()
            tikzplotlib.save(Path(os.path.split(savefile)[0]) / ("tikz_box_plot_" + self.x_label + ".tex"))
        if showfig:
            fig.show()
        return fig, ax

    def save_pandas_csv_result(self, path, step=-1):
        # TODO: Check path
        if step == -1:
            step = max(self.monte_carlo_runs[0].df.step)  # TODO: Assert it's the same for all cases
        cols = ['train loss', 'test loss', 'train accuracy', 'test accuracy']
        for i, run in enumerate(self.monte_carlo_runs):
            df = run.df
            networks_availables = df.network.unique()
            for col, net in enumerate(networks_availables):
                filter = [a == net and b == step for a, b in zip(df.network, df.step)]
                data = df[filter].describe()
                data = data[cols]
                data.to_csv(path + net + "_" + self.x[i] + "_stats.csv")

    def plot_histogram(self, key='test accuracy', step=-1, library='seaborn', showfig=False, savefig=True, title='',
                       extension=".svg"):
        self.monte_carlo_analyzer.plot_histogram(key=key, step=step, library=library, showfig=showfig, savefig=savefig,
                                                 title=title, extension=extension)


class Plotter:

    def __init__(self, path, file_suffix="_results_fit.csv"):
        """
        This class manages the plot of results for a model train.
        It opens the csv files (test and train) saved during training and plots results as wrt each step saved.
        This class is generally used to plot accuracy and loss evolution during training.

        :path: Full path where the csv results are stored
        :file_suffix: (optional) let's you filter csv files to open only files that ends with the suffix.
            By default it opens every csv file it finds.
        """
        if not os.path.exists(path):
            logger.error("Path {} does not exist".format(path))
            sys.exit(-1)
        self.path = Path(path)
        if not file_suffix.endswith(".csv"):
            file_suffix += ".csv"
        self.file_suffix = file_suffix
        self.pandas_list = []
        self.labels = []
        self._csv_to_pandas()

    def _csv_to_pandas(self):
        """
        Opens the csv files as pandas dataframe and stores them in a list (self.pandas_list).
        Also saves the name of the file where it got the pandas frame as a label.
        This function is called by the constructor.
        """
        self.pandas_list = []
        self.labels = []
        files = os.listdir(self.path)
        files.sort()  # Respect the colors for the plot of Monte Carlo.
        # For ComplexVsReal Monte Carlo it has first the Complex model and SECOND the real one.
        # So ordering the files makes sure I open the Complex model first and so it plots with the same colours.
        # TODO: Think a better way without loosing generality (This sort is all done because of the ComplexVsReal case)
        for file in files:
            if file.endswith(self.file_suffix):
                self.pandas_list.append(pd.read_csv(self.path / file))
                self.labels.append(re.sub(self.file_suffix + '$', '', file).replace('_', ' '))

    def reload_data(self):
        """
        If data inside the working path has changed (new csv files or modified csv files),
        this function reloads the data to be plotted with that new information.
        """
        self._csv_to_pandas()

    def get_full_pandas_dataframe(self):
        """
        Merges every dataframe obtained from each csv file into a single dataframe.
        It adds the columns:
            - network: name of the train model
            - step: information of the step index
            - path: path where the information of the train model was saved (used as parameter with the constructor)
        :retun: pd.Dataframe
        """
        # https://pandas.pydata.org/pandas-docs/stable/user_guide/merging.html
        self._csv_to_pandas()
        if len(self.pandas_list) == 0:
            logger.error("Error: There was no csv logs to open")
            sys.exit(-1)
        length = len(self.pandas_list[0])
        for data_frame in self.pandas_list:  # TODO: Check if.
            if not length == len(data_frame):  # What happens if NaN? Can I cope not having same len?
                logger.error("Data frame length should have been {0} and was {1}".format(length, len(data_frame)))

        result = pd.DataFrame({
            'network': [self.get_net_name()] * length,
            # 'step': list(range(length)), # No longer needed, already added in base file
            'path': [self.path] * length
        })

        for data_frame, data_label in zip(self.pandas_list, self.labels):
            # data_frame.columns = [data_label + " " + str(col) for col in data_frame.columns]
            # concatenated = pd.concat(self.pandas_list, keys=self.labels)
            result = pd.concat([result, data_frame], axis=1, sort=False)
        return result

    def get_net_name(self):
        str_to_match = "_metadata.txt"
        for file in os.listdir(self.path):
            if file.endswith(str_to_match):
                # See that there is no need to open the file
                return re.sub(str_to_match + "$", '', file).replace('_', ' ')
        return "Name not found"

    # ====================
    #        Plot
    # ====================

    def plot_everything(self, reload=False, library='plotly', showfig=False, savefig=True, index_loc=None,
                        extension=".svg"):
        if reload:
            self._csv_to_pandas()
        if not len(self.pandas_list) != 0:
            logger.error("Empty pandas list to plot")
            return None
        for key in ["loss", "accuracy"]:
            self.plot_key(key, reload=False, library=library, showfig=showfig, savefig=savefig, index_loc=index_loc,
                          extension=extension)

    def plot_key(self, key='loss', reload=False, library='plotly', showfig=False, savefig=True, index_loc=None,
                 extension=".svg"):
        if reload:
            self._csv_to_pandas()
        if library == 'matplotlib':
            self._plot_matplotlib(key=key, showfig=showfig, savefig=savefig, index_loc=index_loc, extension=extension)
        elif library == 'plotly':
            self._plot_plotly(key=key, showfig=showfig, savefig=savefig, index_loc=index_loc, extension=extension)
        else:
            logger.warning("Unrecognized library to plot " + library)

    def _plot_matplotlib(self, key='loss', showfig=False, savefig=True, index_loc=None, extension=".svg"):
        fig, ax = plt.subplots()
        ax.set_prop_cycle('color', DEFAULT_MATPLOTLIB_COLORS)
        title = None
        for i, data in enumerate(self.pandas_list):
            if title is not None:
                title += " vs. " + self.labels[i]
            else:
                title = self.labels[i]
            for k in data:
                if key in k:
                    if index_loc is not None:
                        if 'stats' in data.keys():
                            data = data[data['stats'] == 'mean']
                        else:
                            logger.warning("Warning: Trying to index an array without index")
                    ax.plot(data[k], 'o-', label=(k.replace(key, '') + self.labels[i]).replace('_', ' '))
        title += " " + key
        fig.legend(loc="upper right")
        ax.set_ylabel(key)
        ax.set_xlabel("step")
        ax.set_title(title)
        if showfig:
            fig.show()
        if savefig:
            fig.savefig(str(self.path / key) + "_matplotlib" + extension, transparent=True)

    def _plot_plotly(self, key='loss', showfig=False, savefig=True, func=min, index_loc=None, extension=".svg"):
        fig = go.Figure()
        annotations = []
        title = None
        for i, data in enumerate(self.pandas_list):
            if title is not None:
                title += " vs. " + self.labels[i]
            else:
                title = self.labels[i]
            j = 0
            for k in data:
                if key in k:
                    color = DEFAULT_PLOTLY_COLORS[j * len(self.pandas_list) + i]
                    j += 1
                    if index_loc is not None:
                        if 'stats' in data.keys():
                            data = data[data['stats'] == 'mean']
                        else:
                            logger.warning("Trying to index an array without index")
                    x = list(range(len(data[k])))
                    fig.add_trace(go.Scatter(x=x, y=data[k], mode='lines',
                                             name=(k.replace(key, '') + self.labels[i]).replace('_', ' '),
                                             line_color=color))
                    # Add points
                    fig.add_trace(go.Scatter(x=[x[-1]],
                                             y=[data[k].to_list()[-1]],
                                             mode='markers',
                                             name='last value',
                                             marker_color=color))
                    # Max/min points
                    func_value = func(data[k])
                    # ATTENTION! this will only give you first occurrence
                    func_index = data[k].to_list().index(func_value)
                    if func_index != len(data[k]) - 1:
                        fig.add_trace(go.Scatter(x=[func_index],
                                                 y=[func_value],
                                                 mode='markers',
                                                 name=func.__name__,
                                                 text=['{0:.2f}%'.format(func_value)],
                                                 textposition="top center",
                                                 marker_color=color))
                        # Min annotations
                        annotations.append(dict(xref="x", yref="y", x=func_index, y=func_value,
                                                xanchor='left', yanchor='middle',
                                                text='{0:.2f}'.format(func_value),
                                                font=dict(family='Arial',
                                                          size=14),
                                                showarrow=False, ay=-40))
                    # Right annotations
                    annotations.append(dict(xref='paper', x=0.95, y=data[k].to_list()[-1],
                                            xanchor='left', yanchor='middle',
                                            text='{0:.2f}'.format(data[k].to_list()[-1]),
                                            font=dict(family='Arial',
                                                      size=16),
                                            showarrow=False))
        title += " " + key
        fig.update_layout(annotations=annotations,
                          title=title,
                          xaxis_title='steps',
                          yaxis_title=key)
        if savefig:
            plotly.offline.plot(fig, filename=str(self.path / key) + ".html",
                                config=PLOTLY_CONFIG, auto_open=showfig)
            # fig.write_image(str(self.path / key) + "_plotly" + extension)
        elif showfig:
            fig.show(config=PLOTLY_CONFIG)


class MonteCarloPlotter(Plotter):

    def __init__(self, path):
        file_suffix = "_statistical_result.csv"
        super().__init__(path, file_suffix=file_suffix)

    def plot_everything(self, reload=False, library='plotly', showfig=False,
                        savefig=True, index_loc='mean', extension=".svg"):
        # Rename this function to change index_loc default value
        super().plot_key(reload=False, library=library, showfig=showfig, savefig=savefig, index_loc=index_loc)

    def plot_key(self, key='test accuracy', reload=False, library='plotly', showfig=False, savefig=True,
                 index_loc='mean', extension=".svg"):
        # Rename this function to change index_loc default value
        super().plot_key(key, reload, library, showfig, savefig, index_loc, extension=extension)

    def plot_line_confidence_interval(self, key='test accuracy', showfig=False, savefig=True, library='matplotlib',
                                      title='', full_border=True, x_axis='step', extension=".svg"):
        if library == 'plotly':
            self._plot_line_confidance_interval_plotly(key=key, showfig=showfig, savefig=savefig,
                                                       title=title, full_border=full_border, x_axis=x_axis)
        elif library == 'matplotlib':
            self._plot_line_confidence_interval_matplotlib(key=key, showfig=showfig, savefig=savefig,
                                                           title=title, x_axis=x_axis, extension=extension)
        else:
            logger.warning("Warning: Unrecognized library to plot " + library)
            return None

    def _plot_line_confidence_interval_matplotlib(self, key='test accuracy', showfig=False, savefig=True,
                                                  title='', x_axis='step', extension=".svg"):
        fig, ax = plt.subplots()
        for i, data in enumerate(self.pandas_list):
            x = data[x_axis].unique().tolist()
            data_mean = data[data['stats'] == 'mean'][key].tolist()
            data_max = data[data['stats'] == 'max'][key].tolist()
            data_min = data[data['stats'] == 'min'][key].tolist()
            data_50 = data[data['stats'] == '50%'][key].tolist()
            data_25 = data[data['stats'] == '25%'][key].tolist()
            data_75 = data[data['stats'] == '75%'][key].tolist()
            ax.plot(x, data_mean, color=DEFAULT_MATPLOTLIB_COLORS[i],
                    label=self.labels[i].replace('_', ' ') + ' mean')
            ax.plot(x, data_50, '--', color=DEFAULT_MATPLOTLIB_COLORS[i],
                    label=self.labels[i].replace('_', ' ') + ' median')
            ax.fill_between(x, data_25, data_75, color=DEFAULT_MATPLOTLIB_COLORS[i], alpha=.4,
                            label=self.labels[i].replace('_', ' ') + ' interquartile')
            ax.fill_between(x, data_min, data_max, color=DEFAULT_MATPLOTLIB_COLORS[i], alpha=.15,
                            label=self.labels[i].replace('_', ' ') + ' border')
        for label in self.labels:
            title += label.replace('_', ' ') + ' vs '
        title = title[:-3] + key

        ax.set_title(title)
        ax.set_xlabel(x_axis)
        ax.set_ylabel(key)
        ax.grid()
        ax.legend()
        if showfig:
            fig.show()
        if savefig:
            os.makedirs(str(self.path / "plots/lines_confidence/"), exist_ok=True)
            fig.savefig(self.path / ("plots/lines_confidence/montecarlo_" +
                                     key.replace(" ", "_") + "_matplotlib" + extension), transparent=True)
            tikzplotlib.save(self.path / ("plots/lines_confidence/montecarlo_" +
                                          key.replace(" ", "_") + "_matplotlib" + ".tex"))
        # set_trace()

    def _plot_line_confidance_interval_plotly(self, key='test accuracy', showfig=False, savefig=True,
                                              title='', full_border=True, x_axis='step'):
        fig = go.Figure()
        for i, data in enumerate(self.pandas_list):
            # set_trace()
            x = data[x_axis].unique().tolist()
            x_rev = x[::-1]
            data_mean = data[data['stats'] == 'mean'][key].tolist()
            data_max = data[data['stats'] == 'max'][key].tolist()
            data_min = data[data['stats'] == 'min'][key][::-1].tolist()
            data_50 = data[data['stats'] == '50%'][key].tolist()
            data_25 = data[data['stats'] == '25%'][key][::-1].tolist()
            data_75 = data[data['stats'] == '75%'][key].tolist()
            # set_trace()
            if full_border:
                fig.add_trace(go.Scatter(
                    x=x + x_rev,
                    y=data_max + data_min,
                    fill='toself',
                    fillcolor=add_transparency(DEFAULT_PLOTLY_COLORS[i], 0.1),
                    line_color=add_transparency(DEFAULT_PLOTLY_COLORS[i], 0),
                    showlegend=True,
                    name=self.labels[i].replace('_', ' ') + " borders",
                ))
            fig.add_trace(go.Scatter(
                x=x + x_rev,
                y=data_75 + data_25,
                fill='toself',
                fillcolor=add_transparency(DEFAULT_PLOTLY_COLORS[i], 0.2),
                line_color=add_transparency(DEFAULT_PLOTLY_COLORS[i], 0),
                showlegend=True,
                name=self.labels[i].replace('_', ' ') + " interquartile",
            ))
            fig.add_trace(go.Scatter(
                x=x, y=data_mean,
                line_color=DEFAULT_PLOTLY_COLORS[i],
                name=self.labels[i].replace('_', ' ') + " mean",
            ))
            fig.add_trace(go.Scatter(
                x=x, y=data_50,
                line=dict(color=DEFAULT_PLOTLY_COLORS[i], dash='dash'),
                name=self.labels[i].replace('_', ' ') + " median",
            ))
        for label in self.labels:
            title += label.replace('_', ' ') + ' vs '
        title = title[:-3] + key

        fig.update_traces(mode='lines')
        fig.update_layout(title=title, xaxis_title=x_axis, yaxis_title=key)

        if savefig:
            os.makedirs(str(self.path / "plots/lines_confidence/"), exist_ok=True)
            plotly.offline.plot(fig,
                                filename=str(self.path / ("plots/lines_confidence/montecarlo_" +
                                                          key.replace(" ", "_"))) + ".html",
                                config=PLOTLY_CONFIG, auto_open=showfig)
            # fig.write_image(str(self.path / ("plots/lines/montecarlo_" + key.replace(" ", "_"))) + extension)
        elif showfig:
            fig.show(config=PLOTLY_CONFIG)

    def plot_train_vs_test(self, key='loss', showfig=False, savefig=True, median=False, x_axis='step'):
        fig = go.Figure()
        # test plots
        label = 'mean'
        if median:
            label = '50%'
        for i, data in enumerate(self.pandas_list):
            x = data[x_axis].unique().tolist()
            data_mean_test = data[data['stats'] == label]["test " + key].tolist()
            fig.add_trace(go.Scatter(
                x=x, y=data_mean_test,
                line_color=DEFAULT_PLOTLY_COLORS[i],
                name=self.labels[i] + " test",
            ))
            data_mean_train = data[data['stats'] == label]["train " + key].tolist()
            fig.add_trace(go.Scatter(
                x=x, y=data_mean_train,
                line_color=DEFAULT_PLOTLY_COLORS[i + len(self.pandas_list)],
                name=self.labels[i].replace("_", " ") + " train ",
            ))
        title = "train and test " + key + " " + label.replace("50%", "median")
        fig.update_traces(mode='lines')
        fig.update_layout(title=title, xaxis_title=x_axis, yaxis_title=key)

        if savefig:
            os.makedirs(self.path / "plots/lines/", exist_ok=True)
            plotly.offline.plot(fig,
                                filename=str(self.path / ("plots/lines/montecarlo_" + key.replace(" ", "_")))
                                         + "_" + label.replace("50%", "median") + ".html",
                                config=PLOTLY_CONFIG, auto_open=showfig)
        elif showfig:
            fig.show(config=PLOTLY_CONFIG)


class MonteCarloAnalyzer:

    def __init__(self, df=None, path=None):
        self.confusion_matrix = []
        if path is not None and df is not None:  # I have data and the place where I want to save it
            self.df = df  # DataFrame with all the data
            self.path = Path(path)
            self.df.to_csv(self.path / "run_data.csv")  # Save the results for latter use
        elif path is not None and df is None:  # Load df from Path
            if not path.endswith('.csv'):
                path += '.csv'
            self.df = pd.read_csv(Path(path))  # Path(__file__).parents[1].absolute() /
            self.path = Path(os.path.split(path)[0])  # Keep only the path and not the filename
        elif path is None and df is not None:  # Save df into default path
            self.path = create_folder("./log/montecarlo/")
            self.df = df  # DataFrame with all the data
            self.df.to_csv(self.path / "run_data.csv")  # Save the results for latter use
        else:  # I have nothing
            self.path = create_folder("./log/montecarlo/")
            self.df = pd.DataFrame()
        self.plotable_info = ['train loss', 'test loss', 'train accuracy', 'test accuracy']  # TODO: Consider delete
        self.monte_carlo_plotter = MonteCarloPlotter(self.path)
        self.summary = []

    def set_df(self, df, conf_mat=None):
        self.df = df  # DataFrame with all the data
        self.df.to_csv(self.path / "run_data.csv")  # Save the results for latter use
        if conf_mat is not None:
            for i in range(len(conf_mat)):
                mat = conf_mat[i]["matrix"]
                group = mat.groupby(mat.index)
                self.confusion_matrix.append({"name": conf_mat[i]["name"], "matrix": group.mean()})
        self.save_stat_results()
        self.monte_carlo_plotter.reload_data()

    def save_stat_results(self):
        # save csv file for each network with 4 columns
        self.summary = []
        networks_availables = self.df.network.unique()
        for net in networks_availables:
            data = self.df[self.df.network == net]
            cols = ['train loss', 'test loss', 'train accuracy', 'test accuracy']
            frames = []
            keys = []
            epochs = []
            for step in data.step.unique():
                desc_frame = data[data.step == step][cols].describe()
                frames.append(desc_frame)
                epochs += [data[data.step == step]['epoch'].iloc[0]] * len(desc_frame)
                keys.append(step)
            data_to_save = pd.concat(frames, keys=keys, names=['step', 'stats'])
            data_to_save['epoch'] = epochs
            data_to_save.to_csv(self.path / (net + "_statistical_result.csv"))
            self.summary.append(data_to_save)
        # Save confusion matrix
        for i in range(len(self.confusion_matrix)):
            self.confusion_matrix[i]["matrix"].to_csv(self.path / (self.confusion_matrix[i]["name"]
                                                                   + "_confusion_matrix.csv"))

    # ------------
    # Plot methods
    # ------------

    def do_all(self, extension=".svg"):
        """self.monte_carlo_plotter.plot_train_vs_test(key='loss')
        self.monte_carlo_plotter.plot_train_vs_test(key='accuracy')
        self.monte_carlo_plotter.plot_train_vs_test(key='loss', median=True)
        self.monte_carlo_plotter.plot_train_vs_test(key='accuracy', median=True)"""

        key_list = ['test accuracy', 'test loss']  # , 'train accuracy', 'train loss']
        for key in key_list:
            # self.plot_3d_hist(key=key)
            for lib in ['seaborn', 'plotly']:
                self.box_plot(key=key, extension=extension, library=lib)
                self.plot_histogram(key=key, library=lib, showfig=False, savefig=True, extension=extension)
                self.monte_carlo_plotter.plot_line_confidence_interval(key=key, x_axis='epoch', library=lib)

    def box_plot(self, step=-1, library='plotly', key='test accuracy', showfig=False, savefig=True, extension='.svg'):
        if library == 'plotly':
            self._box_plot_plotly(key=key, step=step, showfig=showfig, savefig=savefig)
        elif library == 'seaborn':
            self._box_plot_seaborn(key=key, step=step, showfig=showfig, savefig=savefig, extension=extension)
        else:
            logger.warning("Warning: Unrecognized library to plot " + library)
            return None

    def _box_plot_plotly(self, step=-1, key='test accuracy', showfig=False, savefig=True):
        fig = go.Figure()
        if step == -1:
            step = max(self.df.step)
        networks_availables = self.df.network.unique()
        # set_trace()
        for col, net in enumerate(networks_availables):
            filter = [a == net and b == step for a, b in zip(self.df.network, self.df.step)]
            data = self.df[filter]
            fig.add_trace(go.Box(
                y=data[key],
                # x=[self.x[i]] * len(data[key]),
                name=net.replace('_', ' '),
                whiskerwidth=0.2,
                notched=True,  # confidence intervals for the median
                fillcolor=add_transparency(DEFAULT_PLOTLY_COLORS[col], 0.5),
                boxpoints='suspectedoutliers',  # to mark the suspected outliers
                line=dict(color=DEFAULT_PLOTLY_COLORS[col]),
                boxmean=True  # Interesting how sometimes it falls outside the box
            ))
        fig.update_layout(
            # title='Montecarlo Box Plot ' + key,
            xaxis=dict(
                title="network",
            ),
            yaxis=dict(
                title=key,
                autorange=True,
                showgrid=True,
                dtick=0.05,
            ),
            # boxmode='group',
            # boxgroupgap=0,
            # boxgap=0,
            showlegend=False
        )
        if savefig:
            os.makedirs(self.path / "plots/box_plot/", exist_ok=True)
            plotly.offline.plot(fig,
                                filename=str(self.path / (
                                        "plots/box_plot/montecarlo_" + key.replace(" ", "_") + "_box_plot.html")),
                                config=PLOTLY_CONFIG, auto_open=showfig)
            # fig.write_image(str(self.path / ("plots/box_plot/montecarlo_" + key.replace(" ", "_")
            #                                 + "_box_plot" + extension)))
        elif showfig:
            fig.show(config=PLOTLY_CONFIG)

    def _box_plot_seaborn(self, step=-1, key='test accuracy', showfig=False, savefig=True, extension='.svg'):
        if step == -1:
            step = max(self.df.step)
        # Prepare data
        filter = self.df['step'] == step
        data = self.df[filter]

        # Run figure
        fig = plt.figure()
        ax = sns.boxplot(x="network", y=key, data=data, boxprops=dict(alpha=.3))
        # Make black lines the color of the box
        for i, artist in enumerate(ax.artists):
            col = artist.get_facecolor()[:-1]  # the -1 removes the transparency
            artist.set_edgecolor(col)
            for j in range(i * 6, i * 6 + 6):
                line = ax.lines[j]
                line.set_color(col)
                line.set_mfc(col)
                line.set_mec(col)

        if savefig is not None:
            os.makedirs(self.path / "plots/box_plot/", exist_ok=True)
            filename = str(self.path / ("plots/box_plot/montecarlo_" + key.replace(" ", "_") + "_box_plot"))
            fig.savefig(filename + extension)
            # tikzplotlib.clean_figure()
            tikzplotlib.save(filename + ".tex")
        if showfig:
            fig.show()
        return fig, ax

    def show_plotly_table(self):
        # TODO: Not yet debugged
        values = [key for key in self.df.keys()]
        fig = go.Figure(data=[go.Table(
            header=dict(values=list(self.df.columns),
                        fill_color='paleturquoise',
                        align='left'),
            cells=dict(values=[self.df.values.tolist()],
                       fill_color='lavender',
                       align='left'))
        ])
        fig.show(config=PLOTLY_CONFIG)

    def plot_3d_hist(self, steps=None, key='test accuracy', title=''):
        # https://stackoverflow.com/questions/60398154/plotly-how-to-make-a-3d-stacked-histogram/60403270#60403270
        # https://plot.ly/python/v3/3d-filled-line-plots/
        # https://community.plot.ly/t/will-there-be-3d-bar-charts-in-the-future/1045/3
        # https://matplotlib.org/examples/mplot3d/bars3d_demo.html
        if steps is None:
            # steps = [int(x) for x in np.linspace(min(self.df.step), max(self.df.step), 6)]
            steps = [int(x) for x in np.logspace(min(self.df.step), np.log2(max(self.df.step)), 8, base=2)]
            # steps = [int(x) for x in np.logspace(min(self.df.step), np.log10(max(self.df.step)), 8)]
            steps[0] = 0
        networks_availables = self.df.network.unique()
        cols = ['step', key]
        fig = go.Figure()
        for step in steps:  # TODO: verify steps are in df
            for i, net in enumerate(networks_availables):
                filter = [a == net and b == step for a, b in zip(self.df.network, self.df.step)]
                data_to_plot = self.df[filter][cols]
                # https://stackoverflow.com/a/60403270/5931672
                counts, bins = np.histogram(data_to_plot[key], bins=10, density=False)
                counts = list(np.repeat(counts, 2).tolist())  # I do this to stop pycharm warning
                counts.insert(0, 0)
                counts.append(0)
                bins = np.repeat(bins, 2)

                fig.add_traces(go.Scatter3d(x=[step] * len(counts), y=bins, z=counts,
                                            mode='lines', name=net.replace("_", " ") + "; step: " + str(step),
                                            surfacecolor=add_transparency(DEFAULT_PLOTLY_COLORS[i], 0),
                                            # surfaceaxis=0,
                                            line=dict(color=DEFAULT_PLOTLY_COLORS[i], width=4)
                                            )
                               )
                verts, tri = triangulate_histogram([step] * len(counts), bins, counts)
                x, y, z = verts.T
                I, J, K = tri.T
                fig.add_traces(go.Mesh3d(x=x, y=y, z=z, i=I, j=J, k=K, color=DEFAULT_PLOTLY_COLORS[i], opacity=0.4))
        for net in networks_availables:
            title += net + ' '
        title += key + " comparison"
        fig.update_layout(title=title,
                          scene=dict(
                              xaxis=dict(title='step'),
                              yaxis=dict(title=key),
                              zaxis=dict(title='counts'),
                              xaxis_type="log"))
        os.makedirs(self.path / "plots/histogram/", exist_ok=True)
        plotly.offline.plot(fig,
                            filename=str(self.path / (
                                    "plots/histogram/montecarlo_" + key.replace(" ", "_") + "_3d_histogram.html")),
                            config=PLOTLY_CONFIG, auto_open=False)

    def plot_histogram(self, key='test accuracy', step=-1, library='seaborn', showfig=False, savefig=True, title='',
                       extension=".svg"):
        if library == 'matplotlib':
            self._plot_histogram_matplotlib(key=key, step=step, showfig=showfig, savefig=savefig, title=title,
                                            extension=extension)
        elif library == 'plotly':
            self._plot_histogram_plotly(key=key, step=step, showfig=showfig, savefig=savefig, title=title)
        elif library == 'seaborn':
            self._plot_histogram_seaborn(key=key, step=step, showfig=showfig, savefig=savefig, title=title,
                                         extension=extension)
        else:
            logger.warning("Warning: Unrecognized library to plot " + library)
            return None

    def _plot_histogram_matplotlib(self, key='test accuracy', step=-1,
                                   showfig=False, savefig=True, title='', extension=".svg"):
        fig, ax = plt.subplots()
        ax.set_prop_cycle('color', DEFAULT_MATPLOTLIB_COLORS)
        bins = np.linspace(0, 1, 501)
        min_ax = 1.0
        max_ax = 0.0
        networks_availables = self.df.network.unique()
        if step == -1:
            step = max(self.df.step)
        for net in networks_availables:
            filter = [a == net and b == step for a, b in zip(self.df.network, self.df.step)]
            data = self.df[filter]  # Get only the data to plot
            ax.hist(data[key], bins, alpha=0.5, label=net.replace("_", " "))
            min_ax = min(min_ax, min(data[key]))
            max_ax = max(max_ax, max(data[key]))
        title += key + " histogram"
        ax.axis(xmin=min_ax - 0.01, xmax=max_ax + 0.01)
        add_params(fig, ax, x_label=key, y_label="occurances", title=title, loc='upper right',
                   filename=self.path / (
                           "plots/histogram/montecarlo_" + key.replace(" ", "_") + "_matplotlib" + extension),
                   showfig=showfig, savefig=savefig)
        return fig, ax

    def _plot_histogram_plotly(self, key='test accuracy', step=-1, showfig=False, savefig=True, title=''):
        networks_availables = self.df.network.unique()
        if step == -1:
            step = max(self.df.step)
        hist_data = []
        group_labels = []
        for net in networks_availables:
            title += net + ' '
            filter = [a == net and b == step for a, b in zip(self.df.network, self.df.step)]
            data = self.df[filter]  # Get only the data to plot
            hist_data.append(data[key].to_list())
            group_labels.append(net.replace("_", " "))
            # fig.add_trace(px.histogram(np.array(data[key]), marginal="box"))
            # fig.add_trace(go.Histogram(x=np.array(data[key]), name=net))
        fig = ff.create_distplot(hist_data, group_labels, bin_size=0.0005)  # https://plot.ly/python/distplot/
        title += key + " comparison"

        # Overlay both histograms
        fig.update_layout(barmode='overlay')
        # Reduce opacity to see both histograms
        fig.update_traces(opacity=0.75)
        fig.update_layout(title=title.replace('_', ' '),
                          xaxis_title=key)
        if savefig:
            os.makedirs(self.path / "plots/histogram/", exist_ok=True)
            plotly.offline.plot(fig,
                                filename=str(self.path / ("plots/histogram/montecarlo_" + key.replace(" ", "_")
                                                          + "_histogram.html")),
                                config=PLOTLY_CONFIG, auto_open=showfig)
            # fig.write_image(str(self.path / ("plots/histogram/montecarlo_" + key.replace(" ", "_")
            #                                  + "plotly_histogram" + extension)))
        elif showfig:
            fig.show(config=PLOTLY_CONFIG)
        return fig

    def _plot_histogram_seaborn(self, key='test accuracy', step=-1,
                                showfig=True, savefig=True, title='', extension=".svg"):
        fig = plt.figure()
        bins = np.linspace(0, 1, 501)
        min_ax = 1.0
        max_ax = 0.0
        ax = None
        networks_availables = self.df.network.unique()
        # networks_availables = ['complex network', 'real network', 'polar real network']
        if step == -1:
            step = max(self.df.step)
        for net in networks_availables:
            filter = [a == net and b == step for a, b in zip(self.df.network, self.df.step)]
            data = self.df[filter]  # Get only the data to plot
            ax = sns.distplot(data[key], bins, label=net.replace("_", " "))
            min_ax = min(min_ax, min(data[key]))
            max_ax = max(max_ax, max(data[key]))
        title += " " + key + " histogram"
        ax.axis(xmin=min_ax - 0.01, xmax=max_ax + 0.01)
        fig.legend(loc='upper left')  # , bbox_to_anchor=(0., 0.3, 0.5, 0.5))
        add_params(fig, ax, x_label=key.capitalize(), y_label="Occurrences",  # loc='upper left',
                   filename=self.path / (
                           "plots/histogram/montecarlo_" + key.replace(" ", "_") + "_seaborn" + extension),
                   showfig=showfig, savefig=savefig)
        # tikzplotlib.clean_figure()
        tikzplotlib.save(self.path / ("plots/histogram/montecarlo_" + key.replace(" ", "_") + "_seaborn" + ".tex"))
        return fig, ax


if __name__ == "__main__":
    path = "//media/barrachina/data/results/MLSP/results/two hidden layer/TwoLayerTypeA/run_data.csv"
    monte = MonteCarloAnalyzer(path=path)
    # monte.plot_histogram(key='test accuracy', library='seaborn')
    monte.monte_carlo_plotter.plot_line_confidence_interval(key='test loss', x_axis='epochs')

__author__ = 'J. Agustin BARRACHINA'
__version__ = '0.1.26'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'



import numpy as np
import tensorflow as tf
from tensorflow.python.ops import random_ops
from tensorflow.python.ops import stateless_random_ops
from cvnn import logger
import sys
from typing import Union, List, Set, Tuple
from pdb import set_trace
# Initializers:
# https://www.tensorflow.org/api_docs/python/tf/keras/initializers
# https://keras.io/initializers/


class _RandomGenerator(object):
    """
    Random generator that selects appropriate random ops.
    https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/ops/init_ops_v2.py#L1041
    """

    def __init__(self, seed=None):
        super(_RandomGenerator, self).__init__()
        if seed is not None:
            # Stateless random ops requires 2-int seed.
            self.seed = [seed, 0]
        else:
            self.seed = None

    def random_normal(self, shape, mean=0.0, stddev=1, dtype=tf.dtypes.float32):
        """A deterministic random normal if seed is passed."""
        if self.seed:
            op = stateless_random_ops.stateless_random_normal
        else:
            op = random_ops.random_normal
        return op(shape=shape, mean=mean, stddev=stddev, dtype=dtype, seed=self.seed)

    def random_uniform(self, shape, minval, maxval, dtype):
        """A deterministic random uniform if seed is passed."""
        if self.seed:
            op = stateless_random_ops.stateless_random_uniform
        else:
            op = random_ops.random_uniform
        return op(shape=shape, minval=minval, maxval=maxval, dtype=dtype, seed=self.seed)

    def truncated_normal(self, shape, mean, stddev, dtype):
        """A deterministic truncated normal if seed is passed."""
        if self.seed:
            op = stateless_random_ops.stateless_truncated_normal
        else:
            op = random_ops.truncated_normal
        return op(shape=shape, mean=mean, stddev=stddev, dtype=dtype, seed=self.seed)


class RandomInitializer:
    """
    Random initializer helps generate a random tensor of:
        - Either complex or real (floating) data type
        - Either Uniform or Normal distribution
        - Zero mean

    How to use it:
        ```
        # creates a complex tensor of shape (3, 3) distribution with
        #   Re{random_tensor} ~ U[-2, 2] and Im{random_tensor} ~ U[-3, 3]
        random_tensor = RandomInitializer(distribution="uniform")(shape=(3, 3), c_limit=[2, 3], dtype=tf.complex)
        ```
    """
    def __init__(self, distribution="uniform", seed=None):
        """
        :param distribution: It can be either a uniform or a normal distribution.
        :param seed: A Python integer. Used to create a random seed for the distribution. See tf.random.set_seed.
        """
        if distribution.lower() not in {"uniform", "normal"}:
            raise ValueError("Invalid `distribution` argument:", distribution)
        else:
            self.distribution = distribution.lower()
        self._random_generator = _RandomGenerator(seed)     # ATTENTION: I do not check the seed.

    @staticmethod
    def dtype_cast(c_dtype):
        """
        Equivalent to is_floating attribute of `tf.dtypes.Dtype` but casts the input and checks is complex
        (https://www.tensorflow.org/api_docs/python/tf/dtypes/DType)
        """
        c_dtype = tf.as_dtype(c_dtype)
        if not c_dtype.is_complex:
            logger.error("Expecting a complex dtype, got " + str(c_dtype))
            sys.exit(-1)
        return c_dtype.real_dtype

    @staticmethod
    def _compute_fans(shape):
        """
        Taken from https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/ops/init_ops_v2.py#L994
        Computes the number of input and output units for a weight shape.
        Args:
            shape: Integer shape tuple or TF tensor shape.
        Returns:
            A tuple of scalars (fan_in, fan_out).
        """
        if len(shape) < 1:  # Just to avoid errors for constants.
            fan_in = fan_out = 1.
        elif len(shape) == 1:
            fan_in = fan_out = shape[0]
        elif len(shape) == 2:
            fan_in = shape[0]
            fan_out = shape[1]
        else:
            # Assuming convolution kernels (2D, 3D, or more).
            # kernel shape: (..., input_depth, depth)
            receptive_field_size = 1.
            for dim in shape[:-2]:
                receptive_field_size *= dim
            fan_in = shape[-2] * receptive_field_size
            fan_out = shape[-1] * receptive_field_size
        return fan_in, fan_out

    def _call_random_generator(self, shape, arg, dtype):
        if self.distribution == "uniform":
            return self._random_generator.random_uniform(shape=shape, minval=-arg, maxval=arg, dtype=dtype)
        elif self.distribution == "normal":
            # I make this magic number division because that's what tf does on this case
            return self._random_generator.truncated_normal(shape=shape, mean=0.0, stddev=arg / .87962566103423978,
                                                           dtype=dtype)

    def get_random_tensor(self, shape, c_arg=None, r_arg=None, dtype=tf.dtypes.complex64):
        """
        Outputs random values either uniform or normal according to initialization
        :param shape: The shape of the output tensor.
        :param r_arg: Argument.
            If uniform, the output will be a distribution between [-arg, arg].
            If Normal, the output will be a zero-mean gaussian distribution with arg stddev
        :param c_arg: Tuple of the argument for the real and imaginary part respectively.
            *Note: either c_arg or r_arg will be used according to dtype parameter,
                the other will be ignored and can be None.
        :param dtype: The type of the output. Default tf.complex.
        """
        dtype = tf.as_dtype(dtype)
        c_arg, r_arg = self._verify_limits(c_arg, r_arg, dtype)
        if dtype.is_complex:  # Complex layer
            r_dtype = self.dtype_cast(dtype)
            c_arg = tf.cast(c_arg, r_dtype)
            # TODO: I do not yet understand the random_generator thing. I could use tf.random.uniform once I do
            ret = tf.complex(
                self._call_random_generator(shape=shape, arg=c_arg[0], dtype=r_dtype),
                self._call_random_generator(shape=shape, arg=c_arg[1], dtype=r_dtype))
        elif dtype.is_floating:     # Real Layer
            # ret = tf.random.uniform(shape=shape, minval=-limit, maxval=limit, dtype=dtype, seed=seed, name=name)
            r_arg = tf.cast(r_arg, dtype)
            ret = self._call_random_generator(shape=shape, arg=r_arg, dtype=dtype)
        else:
            logger.error("Input_dtype not supported.", exc_info=True)
            sys.exit(-1)
        return ret

    @staticmethod
    def _verify_limits(c_limit, r_limit, dtype):
        if c_limit is None and r_limit is None:
            logger.error("Either one argument must not None")
            sys.exit(-1)
        elif dtype.is_complex and c_limit is None:
            logger.error("complex argument is None but dtype is complex")
            sys.exit(-1)
        elif dtype.is_floating and r_limit is None:
            logger.error("real argument is None but dtype is real")
            sys.exit(-1)
        # TODO: For the moment I do no check the c_limit size 2 and c/r_limit dtype.
        return c_limit, r_limit


class GlorotUniform(RandomInitializer):
    """
    The Glorot uniform initializer, also called Xavier uniform initializer.
    Reference: http://proceedings.mlr.press/v9/glorot10a.html
    Draws samples from a uniform distribution:
        - Real case: `x ~ U[-limit, limit]` where `limit = sqrt(6 / (fan_in + fan_out))`
        - Complex case: `z / Re{z} = Im{z} ~ U[-limit, limit]` where `limit = sqrt(3 / (fan_in + fan_out))`
    where `fan_in` is the number of input units in the weight tensor and `fan_out` is the number of output units.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.GlorotUniform()
    values = initializer(shape=(2, 2))                  # Returns a complex Glorot Uniform tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.GlorotUniform()
    layer = cvnn.layers.Dense(input_size=23, output_size=45, weight_initializer=initializer)
    ```
    """

    __name__ = "Glorot Uniform"

    def __init__(self, seed=None, scale=1.):
        """
        :param seed: A Python integer. An initializer created with a given seed will always produce
            the same random tensor for a given shape and dtype.
        :param scale: Default 1. Scales the limit as `limit = scale * limit`
        """
        if isinstance(scale, float):
            assert scale > 0, "scale must be more than 0. Got " + str(scale)
        else:
            logger.error("scale must be a float. Got " + str(scale))
            sys.exit(-1)
        self.scale = scale
        super(GlorotUniform, self).__init__(distribution="uniform", seed=seed)

    def __call__(self, shape, dtype=tf.dtypes.complex64):
        """
        Returns a tensor object initialized as specified by the initializer.
        :param shape: Shape of the tensor.
        :param dtype: Optional dtype of the tensor. Either floating or complex. ex: tf.complex64 or tf.float32
        """
        fan_in, fan_out = self._compute_fans(shape)
        c_limit = [self.scale * tf.math.sqrt(3. / (fan_in + fan_out)),
                   self.scale * tf.math.sqrt(3. / (fan_in + fan_out))]
        r_limit = self.scale * tf.math.sqrt(6. / (fan_in + fan_out))
        return self.get_random_tensor(shape, c_limit, r_limit, dtype)


class GlorotNormal(RandomInitializer):
    """
    The Glorot normal initializer, also called Xavier normal initializer.
    Reference: http://proceedings.mlr.press/v9/glorot10a.html
        *Note: The reference actually refers to the uniform case but it's analysis was adapted for a normal distribution
    Draws samples from a truncated normal distribution centered on 0 with
     - Real case: `stddev = sqrt(2 / (fan_in + fan_out))`
     - Complex case: real part stddev = complex part stddev = `1 / sqrt(fan_in + fan_out)`
    where `fan_in` is the number of input units in the weight tensor and `fan_out` is the number of output units.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.GlorotNormal()
    values = initializer(shape=(2, 2))                  # Returns a complex Glorot Normal tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.GlorotNormal()
    layer = cvnn.layers.Dense(input_size=23, output_size=45, weight_initializer=initializer)
    ```
    """
    __name__ = "Glorot Normal"

    def __init__(self, seed=None):
        super(GlorotNormal, self).__init__(distribution="normal", seed=seed)

    def __call__(self, shape, dtype=tf.dtypes.complex64):
        """
        Returns a tensor object initialized as specified by the initializer.
        :param shape: Shape of the tensor.
<<<<<<< HEAD
        :param dtype: Optinal dtype of the tensor. Either floating or complex. ex: tf.complex64 or tf.float32
=======
        :param dtype: Optional dtype of the tensor. Either floating or complex. ex: tf.complex63 or tf.float32
>>>>>>> 2a18681b0f0ec5e3a1787b8dd7287f1f3f0de985
        """
        fan_in, fan_out = self._compute_fans(shape)
        c_limit = [tf.math.sqrt(1. / (fan_in + fan_out)), tf.math.sqrt(1. / (fan_in + fan_out))]
        r_limit = tf.math.sqrt(2. / (fan_in + fan_out))
        return self.get_random_tensor(shape, c_limit, r_limit, dtype)


class GlorotUniformCompromise(RandomInitializer):
    """
    The Glorot uniform initializer, also called Xavier uniform initializer.
    Reference: http://proceedings.mlr.press/v9/glorot10a.html
    Draws samples from a uniform distribution:
        - Real case: `x ~ U[-limit, limit]` where `limit = sqrt(6 / (fan_in + fan_out))`
        - Complex case: `z /
            - Re{z} ~ U[-limit, limit]` where `limit = sqrt(3 / (2 * fan_in ))`
            - Im{z} ~ U[-limit, limit]` where `limit = sqrt(3 / ( 2 * fan_out))`
    where `fan_in` is the number of input units in the weight tensor and `fan_out` is the number of output units.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.GlorotUniformCompromise()
    values = initializer(shape=(2, 2))                  # Returns a complex Glorot Uniform tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.GlorotUniformCompromise()
    layer = cvnn.layers.Dense(input_size=23, output_size=45, weight_initializer=initializer)
    ```
    """
    __name__ = "Glorot Uniform Compromise"

    def __init__(self, seed=None):
        super(GlorotUniformCompromise, self).__init__(distribution="uniform", seed=seed)

    def __call__(self, shape, dtype=tf.dtypes.complex64):
        fan_in, fan_out = self._compute_fans(shape)
        c_limit = [tf.math.sqrt(3. / (2. * fan_in)), tf.math.sqrt(3. / (2. * fan_out))]
        r_limit = tf.math.sqrt(6. / (fan_in + fan_out))
        return self.get_random_tensor(shape, c_limit, r_limit, dtype)


class HeNormal(RandomInitializer):
    """
    He normal initializer.
    Reference: https://www.cv-foundation.org/openaccess/content_iccv_2015/html/He_Delving_Deep_into_ICCV_2015_paper.html
    It draws samples from a truncated normal distribution centered on 0 with
        - Real case: `stddev = sqrt(2 / fan_in)`
        - Complex case: real part stddev = complex part stddev = `1 / sqrt(fan_in)`
    where fan_in is the number of input units in the weight tensor.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.HeNormal()
    values = initializer(shape=(2, 2))                  # Returns a complex He Normal tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.HeNormal()
    layer = cvnn.layers.Dense(input_size=23, output_size=45, weight_initializer=initializer)
    ```
    """
    __name__ = "He Normal"

    def __init__(self, seed=None):
        """
        :param seed: A Python integer. An initializer created with a given seed will always
            produce the same random tensor for a given shape and dtype.
        """
        super(HeNormal, self).__init__(distribution="normal", seed=seed)

    def __call__(self, shape, dtype=tf.dtypes.complex64):
        """
        Returns a tensor object initialized as specified by the initializer.
        :param shape: Shape of the tensor.
<<<<<<< HEAD
        :param dtype: Optional dtype of the tensor. Either floating or complex. ex: tf.complex64 or tf.float32
=======
        :param dtype: Optional dtype of the tensor. Either floating or complex. ex: tf.complex63 or tf.float32
>>>>>>> 2a18681b0f0ec5e3a1787b8dd7287f1f3f0de985
        """
        fan_in, fan_out = self._compute_fans(shape)
        c_limit = [tf.math.sqrt(1. / fan_in), tf.math.sqrt(1. / fan_in)]
        r_limit = tf.math.sqrt(2. / fan_in)
        return self.get_random_tensor(shape, c_limit, r_limit, dtype)


class HeUniform(RandomInitializer):
    """
<<<<<<< HEAD
    The He Uniform initializer.
=======
    The Glorot uniform initializer, also called Xavier uniform initializer.
>>>>>>> 2a18681b0f0ec5e3a1787b8dd7287f1f3f0de985
    Reference: http://proceedings.mlr.press/v9/glorot10a.html
    Draws samples from a uniform distribution:
        - Real case: `x ~ U[-limit, limit]` where `limit = sqrt(6 / fan_in)`
        - Complex case: `z / Re{z} = Im{z} ~ U[-limit, limit]` where `limit = sqrt(3 / fan_in)`
    where `fan_in` is the number of input units in the weight tensor.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.HeUniform()
    values = initializer(shape=(2, 2))                  # Returns a complex He Uniform tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.HeUniform()
    layer = cvnn.layers.Dense(input_size=23, output_size=45, weight_initializer=initializer)
    ```
    """
    __name__ = "He Uniform"

    def __init__(self, seed=None):
        super(HeUniform, self).__init__(distribution="uniform", seed=seed)

    def __call__(self, shape, dtype=tf.dtypes.complex64):
        fan_in, fan_out = self._compute_fans(shape)
        c_limit = [tf.math.sqrt(3. / fan_in), tf.math.sqrt(3. / fan_in)]
        r_limit = tf.math.sqrt(6. / fan_in)
        return self.get_random_tensor(shape, c_limit, r_limit, dtype)


class Zeros:
    """
    Creates a tensor with all elements set to zero.

    ```
    > >> cvnn.initializers.Zeros()(shape=(2,2))
    <tf.Tensor: shape=(2, 2), dtype=complex64, numpy=
    array([[0.+0.j, 0.+0.j],
          [0.+0.j, 0.+0.j]], dtype=complex64)>
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.Zeros()
    layer = cvnn.layers.Dense(input_size=23, output_size=45, bias_initializer=initializer)
    ```
    """
    __name__ = "Zeros"

    def __call__(self, shape, dtype=tf.dtypes.complex64):
        return tf.zeros(shape, dtype=dtype)


if __name__ == '__main__':
    # Nothing yet
    set_trace()

__author__ = 'J. Agustin BARRACHINA'
__version__ = '0.0.8'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'



import tensorflow as tf
import sys
import cvnn
import logging

"""
This module contains many complex-valued activation functions to be used by CVNN class.
"""

logger = logging.getLogger(cvnn.__name__)


def apply_activation(act_fun, out):
    """
    Applies activation function `act` to variable `out`
    :param out: Tensor to whom the activation function will be applied
    :param act_fun: function to be applied to out. See the list fo possible activation functions on:
        https://complex-valued-neural-networks.readthedocs.io/en/latest/act_fun.html
    :return: Tensor with the applied activation function
    """
    if act_fun is None:  # No activation function declared
        return out
    elif callable(act_fun):
        if act_fun.__module__ == 'activation_functions' or \
                act_fun.__module__ == 'tensorflow.python.keras.activations':
            return act_fun(out)  # TODO: for the moment is not be possible to give parameters like alpha
        else:
            logger.error("apply_activation Unknown activation function.\n\t"
                         "Can only use activations declared on activation_functions.py or keras.activations")
            sys.exit(-1)
    elif isinstance(act_fun, str):
        try:
            return act_dispatcher[act_fun](out)
        except KeyError:
            logger.warning(str(act_fun) + " is not callable, ignoring it")
        return out


def softmax_real(z, axis=-1):
    """
    Applies the softmax function to the modulus of z.
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Real-valued tensor of the applied activation function
    """
    return tf.keras.activations.softmax(tf.math.abs(z), axis)


# Regression
def linear(z):
    """
    Does not apply any activation function. It just outputs the input.
    :param z: Input tensor variable
    :return: z
    """
    return z


"""
TYPE A: Cartesian form.
"""

# TODO: shall I use tf.nn or tf.keras.activation modules?
# https://stackoverflow.com/questions/54761088/tf-nn-relu-vs-tf-keras-activations-relu
# nn has leaky relu, activation doesn't


def cart_sigmoid(z):
    """
    Applies the function (1.0 / (1.0 + exp(-x))) + j * (1.0 / (1.0 + exp(-y))) where z = x + j * y
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/sigmoid
    :param z: Tensor to be used as input of the activation function
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.sigmoid(tf.math.real(z)),
                              tf.keras.activations.sigmoid(tf.math.imag(z))), dtype=z.dtype)


def cart_elu(z, alpha=1.0):
    """
    Applies the "Exponential linear unit": x if x > 0 and alpha * (exp(x)-1) if x < 0
    To both the real and imaginary part of z.
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/elu
    :param z: Input tensor.
    :param alpha: A scalar, slope of negative section.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.elu(tf.math.real(z), alpha),
                              tf.keras.activations.elu(tf.math.imag(z), alpha)), dtype=z.dtype)


def cart_exponential(z):
    """
    Exponential activation function. Applies to both the real and imag part of z the exponential activation: exp(x)
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/exponential
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.exponential(tf.math.real(z)),
                              tf.keras.activations.exponential(tf.math.imag(z))), dtype=z.dtype)


def cart_hard_sigmoid(z):
    """
    Applies the Hard Sigmoid function to both the real and imag part of z.
    The hard sigmoid function is faster to compute than sigmoid activation.
    Hard sigmoid activation:    0               if x < -2.5
                                1               if x > 2.5
                                0.2 * x + 0.5   if -2.5 <= x <= 2.5
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/hard_sigmoid
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.hard_sigmoid(tf.math.real(z)),
                              tf.keras.activations.hard_sigmoid(tf.math.imag(z))), dtype=z.dtype)


def cart_relu(z, alpha=0.0, max_value=None, threshold=0):
    """
    Applies Rectified Linear Unit to both the real and imag part of z
    The relu function, with default values, it returns element-wise max(x, 0).
    Otherwise, it follows:  f(x) = max_value for x >= max_value,
                            f(x) = x for threshold <= x < max_value,
                            f(x) = alpha * (x - threshold) otherwise.
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/relu
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.relu(tf.math.real(z), alpha, max_value, threshold),
                              tf.keras.activations.relu(tf.math.imag(z), alpha, max_value, threshold)), dtype=z.dtype)


def cart_leaky_relu(z, alpha=0.2, name=None):
    """
    Applies Leaky Rectified Linear Unit to both the real and imag part of z
    https://www.tensorflow.org/api_docs/python/tf/nn/leaky_relu
    http://robotics.stanford.edu/~amaas/papers/relu_hybrid_icml2013_final.pdf
    :param z: Input tensor.
    :param alpha: Slope of the activation function at x < 0. Default: 0.2
    :param name: A name for the operation (optional).
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.nn.leaky_relu(tf.math.real(z), alpha, name),
                              tf.nn.leaky_relu(tf.math.imag(z), alpha, name)), dtype=z.dtype)


def cart_selu(z):
    """
    Applies Scaled Exponential Linear Unit (SELU) to both the real and imag part of z.
    The scaled exponential unit activation: scale * elu(x, alpha).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/selu
    https://arxiv.org/abs/1706.02515
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.selu(tf.math.real(z)),
                              tf.keras.activations.selu(tf.math.imag(z))), dtype=z.dtype)


def cart_softplus(z):
    """
    Applies Softplus activation function to both the real and imag part of z.
    The Softplus function: log(exp(x) + 1)
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softplus
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.softplus(tf.math.real(z)),
                              tf.keras.activations.softplus(tf.math.imag(z))), dtype=z.dtype)


def cart_softsign(z):
    """
    Applies Softsign activation function to both the real and imag part of z.
    The softsign activation: x / (abs(x) + 1).      TODO: typo in tensorflow references (softplus instead of softsign)
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softsign
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.softsign(tf.math.real(z)),
                              tf.keras.activations.softsign(tf.math.imag(z))), dtype=z.dtype)


def cart_tanh(z):
    """
    Applies Hyperbolic Tangent (tanh) activation function to both the real and imag part of z.
    The tanh activation: tanh(x) = sinh(x)/cosh(x) = ((exp(x) - exp(-x))/(exp(x) + exp(-x))).
    The derivative if tanh is computed as 1 - tanh^2 so it should be fast to compute for backprop.
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/tanh
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.tanh(tf.math.real(z)),
                              tf.keras.activations.tanh(tf.math.imag(z))), dtype=z.dtype)


# Classification
def cart_softmax(z, axis=-1):
    """
    Applies the softmax function to both the real and imag part of z.
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.softmax(tf.math.real(z), axis),
                              tf.keras.activations.softmax(tf.math.imag(z), axis)), dtype=z.dtype)


"""
TYPE B: Polar form.
"""
# TODO: for all ReLU functions, the polar form makes no real sense. If we keep the phase because abs(z) > 0


def pol_selu(z):
    """
    Applies Scaled Exponential Linear Unit (SELU) to the absolute value of z, keeping the phase unchanged.
    The scaled exponential unit activation: scale * elu(x, alpha).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/selu
    https://arxiv.org/abs/1706.02515
    :param z: Input tensor.
    :return: Tensor result of the applied activation function

    Logic:
        I must mantain the phase (angle) so: cos(theta) = x_0/r_0 = x_1/r_1.
        For real case, x_0 = r_0 so it also works.
    """
    r_0 = tf.abs(z)
    r_1 = tf.keras.activations.selu(r_0)
    return tf.cast(tf.complex(tf.math.real(z) * r_1 / r_0, tf.math.imag(z) * r_1 / r_0), dtype=z.dtype)


act_dispatcher = {
    'linear': linear,
    'cart_sigmoid': cart_sigmoid,
    'cart_elu': cart_elu,
    'cart_exponential': cart_exponential,
    'cart_hard_sigmoid': cart_hard_sigmoid,
    'cart_relu': cart_relu,
    'cart_leaky_relu': cart_leaky_relu,
    'cart_selu': cart_selu,
    'cart_softplus': cart_softplus,
    'cart_softsign': cart_softsign,
    'cart_tanh': cart_tanh,
    'cart_softmax': cart_softmax,
    'softmax_real': softmax_real,
    'pol_selu': pol_selu
}


__author__ = 'J. Agustin BARRACHINA'
__version__ = '0.0.6'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'



import logging
import colorlog
import re
from cvnn.utils import create_folder

def get_version() -> str:
    VERSIONFILE="_version.py"
    verstrline = open(VERSIONFILE, "rt").read()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        return mo.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# How to comment script header
# https://medium.com/@rukavina.andrei/how-to-write-a-python-script-header-51d3cec13731
__author__ = 'J. Agustin BARRACHINA'
__copyright__ = 'Copyright 2020, {project_name}'
__credits__ = ['{credit_list}']
__license__ = '{license}'
__version__ = get_version()
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'
__status__ = '{dev_status}'

# logging.getLogger('tensorflow').disabled = True     # Removes https://github.com/tensorflow/tensorflow/issues/41557

STRING_FORMATTER = "%(asctime)s — %(levelname)s - %(module)s::%(funcName)s line %(lineno)s — %(message)s"

file_handler = logging.FileHandler(create_folder("./log/logs/") / "logs.log")
formatter = logging.Formatter(STRING_FORMATTER)
file_handler.setFormatter(formatter)

# https://github.com/borntyping/python-colorlog
# https://stackoverflow.com/a/23964880/5931672
console_handler = colorlog.StreamHandler()
console_handler.setFormatter(colorlog.ColoredFormatter('%(log_color)s' + STRING_FORMATTER))

logger = colorlog.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(console_handler)
logger.addHandler(file_handler)






import logging
import cvnn
import cvnn.layers as layers
import cvnn.dataset as dp
from cvnn.cvnn_model import CvnnModel
from cvnn.data_analysis import MonteCarloAnalyzer
from cvnn.layers import Dense
from cvnn.utils import transform_to_real, randomize
from tensorflow.keras.losses import categorical_crossentropy
import pandas as pd
import copy
import sys
from openpyxl import load_workbook, Workbook
from openpyxl.worksheet.table import Table
import os
from tqdm import tqdm
import numpy as np
from pdb import set_trace
from time import sleep

logger = logging.getLogger(cvnn.__name__)


class MonteCarlo:

    def __init__(self):
        self.models = []
        self.pandas_full_data = pd.DataFrame()
        self.confusion_matrix = []
        self.monte_carlo_analyzer = MonteCarloAnalyzer()  # All at None

    def add_model(self, model):
        self.models.append(model)

    def run(self, x, y, data_summary='', polar=False, do_conf_mat=True, validation_split=0.2,
            iterations=100, learning_rate=0.01, epochs=10, batch_size=100,
            shuffle=False, debug=False, display_freq=1, checkpoints=False):
        x, y = randomize(x, y)
        # Reset data frame
        self.pandas_full_data = pd.DataFrame()
        if do_conf_mat:
            for i in range(len(self.models)):
                self.confusion_matrix.append({"name": "model_name", "matrix": pd.DataFrame()})
        self.save_summary_of_run(self._run_summary(iterations, learning_rate, epochs, batch_size, shuffle),
                                 data_summary)
        if not debug:
            pbar = tqdm(total=iterations)
        for it in range(iterations):
            if debug:
                logger.info("Iteration {}/{}".format(it + 1, iterations))
            else:
                pbar.update()
            if shuffle:  # shuffle all data at each iteration
                x, y = randomize(x, y)
            for i, model in enumerate(self.models):
                if model.is_complex():
                    x_fit = x
                else:
                    x_fit = transform_to_real(x, polar=polar)
                test_model = copy.deepcopy(model)
                test_model.fit(x_fit, y, validation_split=validation_split,
                               learning_rate=learning_rate, epochs=epochs, batch_size=batch_size,
                               verbose=debug, display_freq=display_freq,
                               save_csv_history=True)   # Must have save_csv_history to do the montecarlo results latter
                self.pandas_full_data = pd.concat([self.pandas_full_data,
                                                   test_model.plotter.get_full_pandas_dataframe()], sort=False)
                if do_conf_mat:
                    dataset = dp.Dataset(x_fit, y, ratio=1-validation_split)
                    self.confusion_matrix[i]["name"] = test_model.name
                    self.confusion_matrix[i]["matrix"] = pd.concat((self.confusion_matrix[i]["matrix"],
                                                                    test_model.get_confusion_matrix(dataset.x_test,
                                                                                                    dataset.y_test)))
            if checkpoints:
                # Save checkpoint in case Monte Carlo stops in the middle
                self.pandas_full_data.to_csv(self.monte_carlo_analyzer.path / "run_data.csv", index=False)
        if not debug:
            pbar.close()
        self.pandas_full_data = self.pandas_full_data.reset_index(drop=True)
        conf_mat = None
        if do_conf_mat:
            conf_mat = self.confusion_matrix
        self.monte_carlo_analyzer.set_df(self.pandas_full_data, conf_mat)

    @staticmethod
    def _run_summary(iterations, learning_rate, epochs, batch_size, shuffle):
        ret_str = "Monte Carlo run\n"
        ret_str += "\tIterations: {}\n".format(iterations)
        ret_str += "\tepochs: {}\n".format(epochs)
        ret_str += "\tbatch_size: {}\n".format(batch_size)
        ret_str += "\tLearning Rate: {}\n".format(learning_rate)
        if shuffle:
            ret_str += "\tShuffle data at each iteration\n"
        else:
            ret_str += "\tData is not shuffled at each iteration\n"
        return ret_str

    def save_summary_of_run(self, run_summary, data_summary):
        with open(str(self.monte_carlo_analyzer.path / "run_summary.txt"), "w") as file:
            file.write(run_summary)
            file.write(data_summary)
            file.write("Models:\n")
            for model in self.models:
                file.write(model.summary())


class RealVsComplex(MonteCarlo):

    def __init__(self, complex_model, capacity_equivalent=True, equiv_technique='ratio'):
        super().__init__()
        # add models
        self.add_model(complex_model)
        self.add_model(complex_model.get_real_equivalent(capacity_equivalent=capacity_equivalent,
                                                         equiv_technique=equiv_technique, name="real_network"))


# ====================================
#     Excel logging
# ====================================
def run_montecarlo(models, dataset, open_dataset=None, iterations=500,
                   epochs=150, batch_size=100, display_freq=1, learning_rate=0.01,
                   debug=False, polar=False, do_all=True):
    if open_dataset:
        dataset = dp.OpenDataset(open_dataset)  # Warning, open_dataset overwrites dataset

    # Monte Carlo
    monte_carlo = MonteCarlo()
    for model in models:
        monte_carlo.add_model(model)
    if not open_dataset:
        dataset.save_data(monte_carlo.monte_carlo_analyzer.path)
    monte_carlo.run(dataset.x, dataset.y, iterations=iterations, learning_rate=learning_rate,
                    epochs=epochs, batch_size=batch_size, display_freq=display_freq,
                    shuffle=False, debug=debug, data_summary=dataset.summary(), polar=polar)
    if do_all:
        monte_carlo.monte_carlo_analyzer.do_all()

    # Save data to remember later what I did.
    _save_montecarlo_log(path=str(monte_carlo.monte_carlo_analyzer.path),
                         models_names=[str(model.name) for model in models],
                         dataset_name=dataset.dataset_name,
                         num_classes=str(dataset.y.shape[1]),
                         polar_mode='Yes' if polar == 'Apple' else 'No',
                         learning_rate=learning_rate,
                         dataset_size=str(dataset.x.shape[0]),
                         features_size=str(dataset.x.shape[1]), epochs=epochs, batch_size=batch_size
                         )
    return str(monte_carlo.monte_carlo_analyzer.path / "run_data.csv")


def run_gaussian_dataset_montecarlo(iterations=1000, m=10000, n=128, param_list=None,
                                    epochs=150, batch_size=100, display_freq=1, learning_rate=0.01,
                                    shape_raw=None, activation='cart_relu', debug=False, polar=False, do_all=True,
                                    dropout=None):
    # Get parameters
    if param_list is None:
        param_list = [
            [0.5, 1, 1],
            [-0.5, 1, 1]
        ]
    dataset = dp.CorrelatedGaussianCoeffCorrel(m, n, param_list, debug=False)
    mlp_run_real_comparison_montecarlo(dataset, None, iterations, epochs, batch_size, display_freq, learning_rate,
                                       shape_raw, activation, debug, polar, do_all, dropout=dropout)


def mlp_run_real_comparison_montecarlo(dataset, open_dataset=None, iterations=1000,
                                       epochs=150, batch_size=100, display_freq=1, learning_rate=0.01,
                                       shape_raw=None, activation='cart_relu',
                                       debug=False, polar=False, do_all=True, dropout=0.5, validation_split=0.2,
                                       capacity_equivalent=True, equiv_technique='ratio'):
    if shape_raw is None:
        shape_raw = [64]
    if open_dataset:
        dataset = dp.OpenDataset(open_dataset)  # Warning, open_dataset overwrites dataset
    # Create complex network
    input_size = dataset.x.shape[1]  # Size of input
    output_size = dataset.y.shape[1]  # Size of output
    if not len(shape_raw) > 0:
        logger.error("Shape raw was empty")
        sys.exit(-1)
    layers.ComplexLayer.last_layer_output_dtype = None
    layers.ComplexLayer.last_layer_output_size = None
    if len(shape_raw) == 0:
        logger.warning("No hidden layers are used. activation and dropout will be ignored")
        shape = [
            Dense(input_size=input_size, output_size=output_size, activation='softmax_real',
                  input_dtype=np.complex64, dropout=None)
        ]
    else:   # len(shape_raw) > 0:
        shape = [Dense(input_size=input_size, output_size=shape_raw[0], activation=activation,
                       input_dtype=np.complex64, dropout=dropout)]
        for i in range(1, len(shape_raw)):
            shape.append(Dense(output_size=shape_raw[i], activation=activation, dropout=dropout))
        shape.append(Dense(output_size=output_size, activation='softmax_real', dropout=None))

    complex_network = CvnnModel(name="complex_network", shape=shape, loss_fun=categorical_crossentropy,
                                verbose=False, tensorboard=False)

    # Monte Carlo
    monte_carlo = RealVsComplex(complex_network,
                                capacity_equivalent=capacity_equivalent, equiv_technique=equiv_technique)
    sleep(1)  # I have error if not because not enough time passed since creation of models to be in diff folders
    monte_carlo.run(dataset.x, dataset.y, iterations=iterations, learning_rate=learning_rate,
                    epochs=epochs, batch_size=batch_size, display_freq=display_freq,
                    shuffle=False, debug=debug, data_summary=dataset.summary(), polar=polar,
                    validation_split=validation_split)
    if do_all:
        monte_carlo.monte_carlo_analyzer.do_all()

    # Save data to remember later what I did.
    max_epoch = monte_carlo.pandas_full_data['epoch'].max()
    epoch_filter = monte_carlo.pandas_full_data['epoch'] == max_epoch
    complex_filter = monte_carlo.pandas_full_data['network'] == "complex network"
    real_filter = monte_carlo.pandas_full_data['network'] == "real network"
    complex_last_epochs = monte_carlo.pandas_full_data[epoch_filter & complex_filter]
    real_last_epochs = monte_carlo.pandas_full_data[epoch_filter & real_filter]
    complex_median = complex_last_epochs['test accuracy'].median()
    real_median = real_last_epochs['test accuracy'].median()
    _save_rvnn_vs_cvnn_montecarlo_log(path=str(monte_carlo.monte_carlo_analyzer.path),
                                      dataset_name=dataset.dataset_name,
                                      hl=str(len(shape_raw)), shape=str(shape_raw),
                                      dropout=str(dropout), num_classes=str(dataset.y.shape[1]),
                                      polar_mode='Yes' if polar else 'No',
                                      learning_rate=learning_rate, activation=activation,
                                      dataset_size=str(dataset.x.shape[0]),
                                      feature_size=str(dataset.x.shape[1]), epochs=epochs, batch_size=batch_size,
                                      winner='CVNN' if complex_median > real_median else 'RVNN',
                                      complex_median=complex_median, real_median=real_median,
                                      complex_iqr=complex_last_epochs['test accuracy'].quantile(.75)
                                                  - complex_last_epochs['test accuracy'].quantile(.25),
                                      real_iqr=real_last_epochs['test accuracy'].quantile(.75)
                                               - real_last_epochs['test accuracy'].quantile(.25)
                                      )
    return str(monte_carlo.monte_carlo_analyzer.path / "run_data.csv")


# ====================================
#     Excel logging
# ====================================
def _create_excel_file(fieldnames, row_data, filename=None, percentage_cols=None):
    if filename is None:
        filename = './log/montecarlo_summary.xlsx'
    file_exists = os.path.isfile(filename)
    if file_exists:
        wb = load_workbook(filename)
        ws = wb.worksheets[0]
        del ws.tables["Table1"]
    else:
        wb = Workbook()
        ws = wb.worksheets[0]
        ws.append(fieldnames)
    ws.append(row_data)
    # TODO: What if len(row_data) is longer than the dictionary? It corresponds with excel's column names?
    tab = Table(displayName="Table1", ref="A1:" + str(chr(64 + len(row_data))) + str(ws.max_row))
    if percentage_cols is not None:
        for col in percentage_cols:
            ws[col + str(ws.max_row)].number_format = '0.00%'
    ws.add_table(tab)
    wb.save(filename)


def _save_rvnn_vs_cvnn_montecarlo_log(path, dataset_name, hl, shape, dropout, num_classes, polar_mode, learning_rate,
                                      activation,
                                      dataset_size, feature_size, epochs, batch_size, winner,
                                      complex_median, real_median, complex_iqr, real_iqr, comments='', filename=None):
    fieldnames = ['dataset', '# Classes', "Dataset Size", 'Feature Size', "Polar Mode",
                  'HL', 'Shape', 'Dropout', "Learning Rate", "Activation Function", 'epochs', 'batch size',
                  "Winner", "CVNN median", "RVNN median", 'CVNN IQR', 'RVNN IQR',
                  'path', "cvnn version", "Comments"
                  ]
    row_data = [dataset_name, num_classes, dataset_size, feature_size, polar_mode,  # Dataset information
                hl, shape, dropout, learning_rate, activation, epochs, batch_size,  # Model information
                winner, complex_median, real_median, complex_iqr, real_iqr,  # Preliminary results
                path, cvnn.__version__, comments  # Library information
                ]
    percentage_cols = ['N', 'O', 'P', 'Q']
    _create_excel_file(fieldnames, row_data, filename, percentage_cols=percentage_cols)


def _save_montecarlo_log(path, dataset_name, models_names, num_classes, polar_mode, learning_rate, dataset_size,
                         features_size, epochs, batch_size, filename=None):
    fieldnames = [
        'dataset', '# Classes', "Dataset Size", 'Feature Size',  # Dataset information
        'models', 'epochs', 'batch size', "Polar Mode", "Learning Rate",  # Models information
        'path', "cvnn version"  # Library information
    ]
    row_data = [
        dataset_name, num_classes, dataset_size, features_size,
        '-'.join(models_names), epochs, batch_size, polar_mode, learning_rate,
        path, cvnn.__version__
    ]
    _create_excel_file(fieldnames, row_data, filename)


if __name__ == "__main__":
    # Base case with one hidden layer size 64 and dropout 0.5
    run_gaussian_dataset_montecarlo(iterations=10, dropout=None)



import os
import sys
import copy
import re
import logging
import numpy as np
import pandas as pd
from itertools import count  # To count the number of times fit is called
import tensorflow as tf
from datetime import datetime
from pdb import set_trace
from time import strftime, perf_counter, gmtime
from prettytable import PrettyTable
# My own module!
import cvnn.layers as layers
import cvnn.dataset as dp
import cvnn.data_analysis as da
from cvnn.utils import create_folder
from cvnn import logger

try:
    import cPickle as pickle
except ImportError:
    import pickle

VERBOSITY = {0: "SILENT",  # verbosity 0. NADA DE NADA
             2: "FAST",    # verbosity 2. Muestra al final de cada epoch
             4: "PROBAR",  # Shows the progress bar but without acc or loss
             1: "INFO",    # verbosity 1. Muestra entre cada linea
             3: "DEBUG"
             }


def run_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)

    wrapper.has_run = False
    return wrapper


class CvnnModel:
    _fit_count = count(0)  # Used to count the number of layers

    # =====================
    # Constructor and Stuff
    # =====================

    def __init__(self, name, shape, loss_fun, verbose=True, tensorboard=True):
        """
        Constructor
        :param name: Name of the model. It will be used to distinguish models
        :param shape: List of cvnn.layers.ComplexLayer objects
        :param loss_fun: tensorflow.python.keras.losses to be used.
        :param verbose: if True it will print information of the model just created
        :param tensorboard: If true it will save tensorboard information inside log/.../tensorboard_logs/
                - Loss and accuracy
                - Graph
                - Weights histogram
        """
        self.name = name
        # Check all the data is a Layer object
        if not all([isinstance(layer, layers.ComplexLayer) for layer in shape]):
            logger.error("All layers in shape must be a cvnn.layer.Layer", exc_info=True)
            sys.exit(-1)
        self.shape = shape
        self.loss_fun = loss_fun
        self.epochs_done = 0
        self.run_pandas = pd.DataFrame(columns=['step', 'epoch',
                                                'train loss', 'train accuracy', 'test loss', 'test accuracy'])
        if not tf.executing_eagerly():  # Make sure nobody disabled eager execution before running the model
            logging.error("CvnnModel::__init__: TF was not executing eagerly", exc_info=True)
            sys.exit(-1)

        # Folder management for logs
        self.now = datetime.today()
        self.root_dir = create_folder("./log/models/", now=self.now)  # Create folder to store all the information

        # Chekpoints
        self.tensorboard = tensorboard
        # The graph will be saved no matter what. It is only computed once so it won't impact much on the training
        self.graph_writer_logdir = str(self.root_dir.joinpath("tensorboard_logs/graph"))
        self.graph_writer = tf.summary.create_file_writer(self.graph_writer_logdir)
        if self.tensorboard:
            train_writer_logdir = str(self.root_dir.joinpath("tensorboard_logs/train"))
            test_writer_logdir = str(self.root_dir.joinpath("tensorboard_logs/test"))
            weights_writer_logdir = str(self.root_dir.joinpath("tensorboard_logs/weights"))
            activation_writer_logdir = str(self.root_dir.joinpath("tensorboard_logs/activation"))
            gradients_writer_logdir = str(self.root_dir.joinpath("tensorboard_logs/gradients"))
            self.train_summary_writer = tf.summary.create_file_writer(train_writer_logdir)
            self.test_summary_writer = tf.summary.create_file_writer(test_writer_logdir)
            self.weights_summary_writer = tf.summary.create_file_writer(weights_writer_logdir)
            self.activation_summary_writer = tf.summary.create_file_writer(activation_writer_logdir)
            self.gradients_summary_writer = tf.summary.create_file_writer(gradients_writer_logdir)

        # print("Saving {}/{}_metadata.txt".format(self.root_dir, self.name))     # To debug the warning message
        self._manage_string(self.summary(), self._get_verbose(verbose), filename=self.name + "_metadata.txt", mode="x")
        self.plotter = da.Plotter(self.root_dir)

    def call(self, x):
        """
        Forward result of the network
        :param x: Data input to be calculated
        :return: Output of the netowrk
        """
        for i in range(len(self.shape)):  # Apply all the layers
            x = self.shape[i].call(x)
        return x

    def _apply_loss(self, y_true, y_pred):
        """
        Private! Use "evaluate loss" instead
        """
        # TODO: This can actually be static and give the parameter of loss_fun?
        if callable(self.loss_fun):
            if self.loss_fun.__module__ != 'tensorflow.python.keras.losses':
                logger.error("Unknown loss function.\n\t "
                             "Can only use losses declared on tensorflow.python.keras.losses", exc_info=True)
                sys.exit(-1)
        return tf.reduce_mean(input_tensor=self.loss_fun(y_true, y_pred), name=self.loss_fun.__name__)

    def is_complex(self):
        """
        :return: True if the network is complex. False otherwise.
        """
        dtype = self.shape[0].get_input_dtype()
        if dtype == np.complex64 or dtype == np.complex128:
            return True
        else:
            return False

    # ====================
    #     Copy methods
    # ====================

    def __deepcopy__(self, memodict=None):
        """
        This function is used to create a copy of the model.
        Used for the Monte Carlo simulation. Creates a copy of the model and then trains them.
        ATTENTION: This does not keep the model's weights but randomly initializes them.
            (makes sense like that for the Monte Carlo simulation)
        :param memodict:
        :return: A copy of the current model
        """
        if memodict is None:
            memodict = {}
        new_shape = []
        layers.ComplexLayer.last_layer_output_dtype = None
        layers.ComplexLayer.last_layer_output_size = None
        for layer in self.shape:
            if isinstance(layer, layers.ComplexLayer):
                new_shape.append(copy.deepcopy(layer))
            else:
                logger.error("Layer " + str(layer) + " not child of cvnn.layers.ComplexLayer")
                sys.exit(-1)
        return CvnnModel(self.name, new_shape, self.loss_fun, verbose=False, tensorboard=self.tensorboard)

    def _get_real_equivalent_multiplier(self, classifier: bool = True, capacity_equivalent: bool = True,
                                        equiv_technique: str = 'ratio'):
        """
        Returns an array (output_multiplier) of size self.shape (number of hidden layers + output layer)
            one must multiply the real valued equivalent layer
        In other words, the real valued equivalent layer 'i' will have:
            neurons_real_valued_layer[i] = output_multiplier[i] * neurons_complex_valued_layer[i]
        :param classifier: Boolean (default = True) weather the model's task is a to classify (True) or
                                                                                            a regression task (False)
        :param capacity_equivalent: An equivalent model can be equivalent in terms of layer neurons or
                        trainable parameters (capacity equivalent according to (https://arxiv.org/abs/1811.12351)
            - True, it creates a capacity-equivalent model in terms of trainable parameters
            - False, it will double all layer size (except the last one if classifier=True)
        :param equiv_technique: Used to define the strategy of the capacity equivalent model.
            This parameter is ignored if capacity_equivalent=False
            - 'ratio': neurons_real_valued_layer[i] = r * neurons_complex_valued_layer[i], 'r' constant for all 'i'
            - 'alternate': Method described in https://arxiv.org/abs/1811.12351 where one alternates between
                    multiplying by 2 or 1. Special case on the middle is treated as a compromise between the two.
        :return: output_multiplier
        """
        if capacity_equivalent:
            if equiv_technique == "alternate":
                output_multiplier = self._get_alternate_capacity_equivalent(classifier)
            elif equiv_technique == "ratio":
                output_multiplier = self._get_ratio_capacity_equivalent(classifier)
            else:
                logger.error("Unknown equiv_technique " + equiv_technique)
                sys.exit(-1)
        else:
            output_multiplier = 2 * np.ones(len(self.shape)).astype(int)
            if classifier:
                output_multiplier[-1] = 1
        return output_multiplier

    def _get_ratio_capacity_equivalent(self, classification: bool = True, bias_adjust: bool = True):
        """
        Generates output_multiplier keeping not only the same capacity but keeping a constant ratio between the
                                                                                                        model's layers
        This helps keeps the 'aspect' or shape of the model my making:
            neurons_real_layer_i = ratio * neurons_complex_layer_i
        :param classification: True (default) if the model is a classification model. False otherwise.
        :param bias_adjust: True (default) if taking into account the bias as a trainable parameter. If not it will
            only match the real valued parameters of the weights
        """
        model_in_c = self.shape[0].input_size
        model_out_c = self.shape[-1].output_size
        x_c = [self.shape[i].output_size for i in range(len(self.shape[:-1]))]
        p_c = np.sum([2 * x.input_size * x.output_size for x in self.shape])  # real valued complex trainable params
        if bias_adjust:
            p_c = p_c + 2 * np.sum(x_c) + 2 * model_out_c
        model_in_r = 2 * model_in_c
        model_out_r = model_out_c if classification else 2 * model_out_c
        # Quadratic equation
        if len(x_c) > 1:
            quadratic_c = -p_c
            quadratic_b = model_in_r * x_c[0] + model_out_r
            if bias_adjust:
                quadratic_b = quadratic_b + np.sum(x_c) + model_out_c
            quadratic_a = np.sum([x_c[i] * x_c[i + 1] for i in range(len(x_c) - 1)])

            ratio = (-quadratic_b + np.sqrt(quadratic_b ** 2 - 4 * quadratic_c * quadratic_a)) / (2 * quadratic_a)
            # The result MUST be positive so I use the '+' solution
            if not 1 <= ratio <= 2:
                logger.error("Ratio {} has a weird value. This function must have a bug.".format(ratio))
        else:
            ratio = 2 * (model_in_c + model_out_c) / (model_in_r + model_out_r)     # TODO: Verify
        return [ratio] * len(x_c) + [1 if classification else 2]

    def _get_alternate_capacity_equivalent(self, classification: bool = True):
        """
        Generates output_multiplier using the alternate method described in https://arxiv.org/abs/1811.12351 which
            doubles or not the layer if it's neighbor was doubled or not (making the opposite).
        The code fills output_multiplier from both senses:
            output_multiplier = [ ... , .... ]
                          --->     <---
        If when both ends meet there's not a coincidence (example: [..., 1, 1, ...]) then
            the code will find a compromise between the two to keep the same real valued trainable parameters.
        """
        output_multiplier = np.zeros(len(self.shape) + 1)
        output_multiplier[0] = 2
        output_multiplier[-1] = 1 if classification else 2
        i: int = 1
        while i <= len(self.shape) - i:
            output_multiplier[i] = 2 if output_multiplier[i - 1] == 1 else 1  # From beginning
            output_multiplier[-1 - i] = 2 if output_multiplier[-i] == 1 else 1  # From the end
            if i == len(self.shape) - i and output_multiplier[i - 1] != output_multiplier[i + 1] or \
                    i + 1 == len(self.shape) - i and output_multiplier[i] == output_multiplier[i + 1]:
                m_inf = self.shape[i - 1].output_size
                m_sup = self.shape[i + 1].output_size
                output_multiplier[i] = 2 * (m_inf + m_sup) / (m_inf + 2 * m_sup)
            i += 1
        return output_multiplier[1:]

    def get_real_equivalent(self, classifier: bool = True, capacity_equivalent: bool = True,
                            equiv_technique: str = 'ratio', name: str = None):
        """
        Creates a new model equivalent of current model. If model is already real throws and error.
        :param classifier: True (default) if the model is a classification model. False otherwise.
        :param capacity_equivalent: An equivalent model can be equivalent in terms of layer neurons or
                        trainable parameters (capacity equivalent according to: https://arxiv.org/abs/1811.12351)
            - True, it creates a capacity-equivalent model in terms of trainable parameters
            - False, it will double all layer size (except the last one if classifier=True)
        :param equiv_technique: Used to define the strategy of the capacity equivalent model.
            This parameter is ignored if capacity_equivalent=False
            - 'ratio': neurons_real_valued_layer[i] = r * neurons_complex_valued_layer[i], 'r' constant for all 'i'
            - 'alternate': Method described in https://arxiv.org/abs/1811.12351 where one alternates between
                    multiplying by 2 or 1. Special case on the middle is treated as a compromise between the two.
        :param name: name of the new network to be created.
            If None (Default) it will use same name as current model with "_real_equiv" suffix
        :return: CvnnModel() real equivalent model
        """
        if not self.is_complex():
            logger.error("model {} was already real".format(self.name))
            sys.exit(-1)
        equiv_technique = equiv_technique.lower()
        if equiv_technique not in {"ratio", "alternate"}:
            logger.error("Invalid `equivalent_technique` argument: " + equiv_technique)
            sys.exit(-1)
        # assert len(self.shape) != 0
        real_shape = []
        layers.ComplexLayer.last_layer_output_dtype = None
        layers.ComplexLayer.last_layer_output_size = None
        output_multiplier = self._get_real_equivalent_multiplier(classifier, capacity_equivalent, equiv_technique)
        for i, layer in enumerate(self.shape):
            if isinstance(layer, layers.ComplexLayer):
                if isinstance(layer, layers.Dense):  # TODO: Check if I can do this with kargs or sth
                    real_shape.append(layer.get_real_equivalent(
                        output_multiplier=output_multiplier[i],
                        input_multiplier=output_multiplier[i - 1] if i > 0 else 2))
                else:
                    real_shape.append(layer.get_real_equivalent())
            else:
                sys.exit("Layer " + str(layer) + " unknown")
        if name is None:
            name = self.name + "_real_equiv"
        # set_trace()
        return CvnnModel(name=name, shape=real_shape, loss_fun=self.loss_fun,
                         tensorboard=self.tensorboard, verbose=False)

    # ====================
    #          Train
    # ====================

    @run_once
    def _start_graph_tensorflow(self):
        # https://github.com/tensorflow/agents/issues/162#issuecomment-512553963
        # Bracket the function call with
        # tf.summary.trace_on() and tf.summary.trace_export().
        # https://stackoverflow.com/questions/4103773/efficient-way-of-having-a-function-only-execute-once-in-a-loop
        tf.summary.trace_on(graph=True, profiler=True)  # https://www.tensorflow.org/tensorboard/graphs

    @run_once
    def _end_graph_tensorflow(self):
        with self.graph_writer.as_default():
            tf.summary.trace_export(name="graph", step=0, profiler_outdir=self.graph_writer_logdir)

    # Add '@tf.function' to accelerate the code by a lot!
    @tf.function
    def _train_step(self, x_train_batch, y_train_batch, learning_rate):
        """
        Performs one step of the training
        :param x_train_batch: input
        :param y_train_batch: labels
        :param learning_rate: learning rate fot the gradient descent
        :return: None
        """
        with tf.GradientTape() as tape:
            with tf.name_scope("Forward_Phase"):
                x_called = self.call(x_train_batch)  # Forward mode computation
            # Loss function computation
            with tf.name_scope("Loss"):
                current_loss = self._apply_loss(y_train_batch, x_called)  # Compute loss

        # Calculating gradient
        with tf.name_scope("Gradient"):
            variables = []
            for lay in self.shape:
                variables.extend(lay.trainable_variables())  # TODO: Debug this for all layers.
            gradients = tape.gradient(current_loss, variables)  # Compute gradients
            assert all(g is not None for g in gradients)

        # Backpropagation
        with tf.name_scope("Optimizer"):
            for i, val in enumerate(variables):
                val.assign(val - learning_rate * gradients[i])  # TODO: For the moment the optimization is only GD

    def fit(self, x, y=None,
            validation_split=0.0, validation_data=None,
            learning_rate: float = 0.01, epochs: int = 10, batch_size: int = 32,
            verbose=True, display_freq: int = 1,
            save_model_checkpoints=False, save_csv_history=True, shuffle=True):
        """
        Trains the model for a fixed number of epochs (iterations on a dataset).

        :param x: Input data. It could be:
            - A Numpy array (or array-like), or a list of arrays (in case the model has multiple inputs).
            - A TensorFlow tensor, or a list of tensors (in case the model has multiple inputs).
            - A tf.data dataset. Should return a tuple (inputs, targets). Preferred data type (less overhead).
        :param y: Labels/Target data. Like the input data x, it could be either Numpy array(s) or TensorFlow tensor(s).
            If f x is a dataset then y will be ignored (default None)
        :param validation_split: Float between 0 and 1.
            Percentage of the input data to be used as test set (the rest will be use as train set)
            Default: 0.0 (No validation set).
            This input is ignored if validation_data is given.
        :param validation_data: Data on which to evaluate the loss and any model metrics at the end of each epoch.
            The model will not be trained on this data. This parameter takes precedence over validation_split.
            It can be:
                - tuple (x_val, y_val) of Numpy arrays or tensors. Preferred data type (less overhead).
                - A tf.data dataset.
        :param learning_rate: Learning rate for the gradient descent. For the moment only SGD is supported.
        :param epochs: (uint) Number of epochs to do.
        :param batch_size: (uint) Batch size of the data. Default 32 (because keras use 32 so... why not?)
        :param verbose: Verbosity Mode
            It can be:
                - Bool: False defaults to 0 and True to 1.
                - Int
                - String: Matching the modes string
            Verbosity Modes:
                - "SILENT" or 0:  No prints of any kind
                - "FAST" or 2:    Does not show the progress bar of each epoch.
                    Verbosity modes "FAST" and "SILENT" saves the csv file (if save_csv_history) less often.
                    Making it faster riskier of data loss
                - "PROBAR" or 4:  Shows progress bar but does not show accuracy or loss (helps on speed)
                - "INFO" or 1:    Shows a progress bar with current accuracy and loss
                - "DEBUG" or 3:   Shows start and end messages and also the progress bar with current accuracy and loss
            Verbosity modes 0, 1 and 2 are coincident with tensorflow's fit verbose parameter:
                https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit
        :param display_freq: Integer (Default 1)
            Frequency on terms of epochs before saving information and running a checkpoint.
        :param save_model_checkpoints: Save the model to be able to load and continue training later (Not yet working)
        :param save_csv_history: Save information of the train and test loss and accuracy on csv files.
        :param shuffle: (Boolean) Whether to shuffle the training data before each epoch. Default: True
        :return: None
        """
        # Check input
        verbose = self.verify_fit_input(save_model_checkpoints, epochs, batch_size, learning_rate, display_freq, verbose)
        # Prepare dataset
        train_dataset, (x_test, y_test) = self._process_dataset(x, y, validation_split, validation_data)
        train_dataset = train_dataset.batch(batch_size=batch_size)  # TODO: Check if batch_size = 1
        # Print start condition
        start_status = ''
        if x_test is not None and y_test is not None:
            start_status = self._get_str_evaluate(self.epochs_done, epochs, x_test, y_test)
        self._manage_string("Starting training...\nLearning rate = " + str(learning_rate) + "\n" +
                            "Epochs = " + str(epochs) + "\nBatch Size = " + str(batch_size) + "\n" +
                            start_status, verbose)
        # -----------------------------------------------------
        # input processing ended
        # num_tr_iter = int(x_train.shape[0] / batch_size)  # Number of training iterations in each epoch
        epochs_before_fit = self.epochs_done
        start_time = perf_counter()
        total_iteration = None
        for epoch in range(epochs):
            iteration = 0
            if verbose in ("FAST", "INFO", "DEBUG", "PROBAR"):
                tf.print("Epoch {0}/{1}".format(self.epochs_done + 1, epochs))
                progbar = tf.keras.utils.Progbar(total_iteration,
                                                 stateful_metrics=[('loss', 0), ('accuracy', 0),
                                                                   ('val_loss', 0), ('val_accuracy', 0)])
            # Randomly shuffle the training data at the beginning of each epoch
            if shuffle:
                train_dataset = train_dataset.shuffle(buffer_size=50000)
            for x_batch, y_batch in train_dataset.prefetch(tf.data.experimental.AUTOTUNE).cache():
                if verbose in ("INFO", "DEBUG", "PROBAR"):
                    values = None
                    if verbose != "PROBAR":
                        train_loss, train_acc = self.evaluate_train_and_test(x_batch, y_batch)[:2]
                        values = [('loss', train_loss), ('accuracy', train_acc)]
                    progbar.update(iteration, values=values)
                iteration += 1
                # Run optimization op (backpropagation)
                self._start_graph_tensorflow()
                self._train_step(x_batch, y_batch, learning_rate)
                self._end_graph_tensorflow()
            if total_iteration is None:
                total_iteration = iteration
            if verbose in ("FAST", "INFO", "DEBUG", "PROBAR"):
                values = None
                if verbose != "PROBAR":
                    train_loss, train_acc, test_loss, test_acc = self.evaluate_train_and_test(x_batch, y_batch,
                                                                                              x_test,  y_test)
                    values = [('loss', train_loss), ('accuracy', train_acc)]
                    if test_loss is not None:
                        values += [('val_loss', test_loss), ('val_accuracy', test_acc)]
                progbar.update(iteration,
                               values=values,
                               finalize=True)
            # Save checkpoint if needed
            if (epochs_before_fit + epoch) % display_freq == 0:
                self._run_checkpoint(x_batch, y_batch, x_test, y_test,  # Shall I use batch to be more efficient?
                                     step=(epochs_before_fit + epoch) * total_iteration + total_iteration,
                                     num_tr_iter=total_iteration, total_epochs=epochs_before_fit + epochs,
                                     verbose="SILENT", save_model_checkpoints=save_model_checkpoints,
                                     save_csv_checkpoints=save_csv_history,
                                     fast_mode=True if verbose in ("SILENT", "FAST") else False)
            self.epochs_done += 1

        # After epochs
        end_time = perf_counter()
        x_train = x_batch
        y_train = y_batch
        self._run_checkpoint(x_train, y_train, x_test, y_test,
                             step=epochs * total_iteration + total_iteration, num_tr_iter=total_iteration,
                             total_epochs=epochs_before_fit + epochs,
                             fast_mode=False)       # I use false to save the csv file this time
        end_status = self._get_str_evaluate(epochs, epochs, x_train, y_train, x_test, y_test)
        self._manage_string("Train finished...\n" +
                            end_status +
                            "\nTraining time: {}s".format(strftime("%H:%M:%S", gmtime(end_time - start_time))),
                            verbose)
        self.plotter.reload_data()

    def verify_fit_input(self, save_model_checkpoints, epochs: int, batch_size: int,
                         learning_rate: float, display_freq: int, verbose):
        assert not save_model_checkpoints  # TODO: Not working for the moment, sorry!
        if not (isinstance(epochs, int) and epochs > 0):
            logger.error("Epochs must be unsigned integer", exc_info=True)
            sys.exit(-1)
        if not (isinstance(batch_size, int) and batch_size > 0):
            logger.error("Batch size must be unsigned integer", exc_info=True)
            sys.exit(-1)
        if not learning_rate > 0:
            logger.error("Learning rate must be positive", exc_info=True)
            sys.exit(-1)
        if isinstance(display_freq, int):
            assert display_freq > 0, "display_freq must be positive"
        else:
            logger.error("display_freq must be a unsigned integer. Got" + str(display_freq))
            sys.exit(-1)
        return self._get_verbose(verbose)

    """
    @classmethod
    def _extractfromlocal(cls, model):  # extracts attributes from the local thrading container
        model._thread_local = model._thread_local.__dict__
        for attr in model.__dict__.values():
            if '_thread_local' in dir(attr):
                cls._extractfromlocal(attr)

    @classmethod
    def _loadtolocal(cls, model):   # puts attributes back to the local threading container
        aux=threading.local()
        aux.__dict__.update(model._thread_local)
        model._thread_local = aux
        for attr in model.__dict__.values():
            if '_thread_local' in dir(attr):
                cls._loadtolocal(attr)
    """

    @staticmethod
    def _get_verbose(verbose):
        if isinstance(verbose, bool):
            return "INFO" if verbose else "SILENT"
        elif isinstance(verbose, int):
            return VERBOSITY[verbose]
        elif isinstance(verbose, str):
            if verbose.upper() in VERBOSITY.values():
                return verbose.upper()
            else:
                supported_strings = "\n\t"
                for s in VERBOSITY.values():
                    supported_strings += s + "\n\t"
                logger.error("verbose string unknown \"" + verbose + "\". Supported: " + supported_strings)
                sys.exit(-1)
        else:
            logger.error("verbose datatype (" + str(type(verbose)) + ") not supported")

    @staticmethod
    def _process_dataset(x, y, validation_split=0.0, validation_data=None):
        test_dataset = None
        if isinstance(x, (list, tuple, np.ndarray)):
            if validation_data is None:
                assert 0 <= validation_split < 1, "Ratio should be between [0, 1)"
                dataset_length = np.shape(x)[0]
                x_train = x[int(dataset_length * validation_split):]
                y_train = y[int(dataset_length * validation_split):]
                x_test = x[:int(dataset_length * validation_split)]
                y_test = y[:int(dataset_length * validation_split)]
                if len(x_test) != 0:
                    validation_data = (np.array(x_test), np.array(y_test))
                train_dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train))
            else:
                if validation_split != 0:
                    logger.warning("validation_split was given but will be ignored because "
                                   "validation_data was not None")
                train_dataset = tf.data.Dataset.from_tensor_slices((x, y))
        elif isinstance(x, tf.data.Dataset):
            if y is not None:
                logger.warning("y is ignored because x was a Dataset (and should contain the labels), "
                               "however, y was not None")
            train_dataset = x.unbatch()
        else:
            logger.error("dataset type ({}) not supported".format(type(x)))
            sys.exit(-1)
        if validation_data is not None:
            if isinstance(validation_data, (list, tuple, np.ndarray)):
                assert len(validation_data) == 2, \
                    "validation_data must have size 2. Size was {}".format(len(validation_data))
                test_dataset = np.array(validation_data[0]), np.array(validation_data[1])
            elif isinstance(validation_data, tf.data.Dataset):
                x_test = []
                y_test = []
                for element in validation_data.unbatch():
                    x_test.append(element[0].numpy())
                    y_test.append(element[1].numpy())
                test_dataset = [np.array(x_test), np.array(y_test)]
            else:
                logger.error("validation_data type ({}) not supported ".format(type(validation_data)))
                sys.exit(-1)
        assert isinstance(test_dataset, (list, tuple, np.ndarray)) or test_dataset is None
        if test_dataset is None:
            test_dataset = (None, None)
        return train_dataset, test_dataset

    @classmethod  # https://stackoverflow.com/a/2709848/5931672
    def loader(cls, f):
        return pickle.load(f)  # TODO: not yet tested

    # ==========================
    # Predict models and results
    # ==========================

    def predict(self, x):
        """
        Predicts the value of the class.
        ATTENTION: Use this only for classification tasks. For regression use "call" method.
        :param x: Input
        :return: Prediction of the class that x belongs to.
        """
        y_out = self.call(x)
        return tf.math.argmax(y_out, 1)

    def evaluate_loss(self, x, y):
        """
        Computes the output of x and computes the loss using y
        :param x: Input of the netwotk
        :param y: Labels
        :return: loss value
        """
        return self._apply_loss(y, self.call(x)).numpy()

    def evaluate_accuracy(self, x, y):
        """
        Computes the output of x and returns the accuracy using y as labels
        :param x: Input of the netwotk
        :param y: Labels
        :return: accuracy
        """
        y_pred = self.predict(x)
        if y_pred.shape == y.shape:
            y_labels = y
        else:
            y_labels = tf.math.argmax(y, 1)
        return tf.math.reduce_mean(tf.dtypes.cast(tf.math.equal(y_pred, y_labels), tf.float64)).numpy()

    def evaluate(self, x, y):
        """
        Compues both the loss and accuracy using "evaluate_loss" and "evaluate_accuracy"
        :param x: Input of the netwotk
        :param y: Labels
        :return: tuple (loss, accuracy)
        """
        return self.evaluate_loss(x, y), self.evaluate_accuracy(x, y)

    def evaluate_train_and_test(self, train_x, train_y, test_x=None, test_y=None):
        train_loss, train_acc = self.evaluate(train_x, train_y)
        test_loss, test_acc = None, None
        if test_x is not None and test_y is not None:
            test_loss, test_acc = self.evaluate(test_x, test_y)
        return train_loss, train_acc, test_loss, test_acc

    def get_confusion_matrix(self, x, y, save_result=False):
        """
        Generates a pandas data-frame with the confusion matrix of result of x and y (labels)
        :param x: data to which apply the model
        :param y: labels
        :param save_result: if True it will save the confusion matrix as a csv at models path
        :return: Confusion matrix pandas data-frame
        """
        filename = None
        if save_result:
            filename = self.root_dir / "categorical.csv"
        return da.confusion_matrix(self.call(x), y, filename=filename)

    # ===========
    # Checkpoints
    # ===========

    def _run_checkpoint(self, x_train, y_train, x_test=None, y_test=None,
                        step=0, num_tr_iter=0, total_epochs=0, verbose="SILENT",
                        save_model_checkpoints=False, save_csv_checkpoints=True, fast_mode=False):
        """
        Saves whatever needs to be saved (tensorboard, csv of train and test acc and loss, model weigths, etc.
        :param x_train: Train data
        :param y_train: Tran labels
        :param x_test: Test data (optional)
        :param y_test: Test labels (optional)
        :param step: step of the training.
        :param num_tr_iter: Total number of iterations per epoch.
        :param total_epochs: Total epochs to be done on the training.
        :param fast_mode: if True it will save the loss and accuracy as a csv on each file.
                This will prevent the loss of data but will make the training longer.
                Default: False.
        :param verbose: Print the results on console to visualize the training step.
        :param save_model_checkpoints: Save the model to be able to load and continue training later (Not yet working)
        :param save_csv_checkpoints: Save information of the train and test loss and accuracy on csv files.
        :return: None
        """
        # First I check if at least one is needed. If not better don't compute the information.
        if save_csv_checkpoints or save_model_checkpoints or verbose:
            train_loss, train_acc, test_loss, test_acc = self.evaluate_train_and_test(x_train, y_train, x_test, y_test)
        if self.tensorboard:  # Save tensorboard data
            self._tensorboard_checkpoint(x_train, y_train, x_test, y_test)
        if save_csv_checkpoints:
            # With fast mode False I save the checkpoint in a csv.
            # It will take longer to run because I create a file each time
            # but if I don't do it and something happens I will loose all the information
            self._save_current_loss_and_acc(self.name + '_results_fit',
                                            train_loss, train_acc, test_loss, test_acc, step, fast_mode=fast_mode)
        if save_model_checkpoints:  # Save model weights
            self.save(test_loss, test_acc)
        if verbose != "SILENT":  # I first check if it makes sense to get the str
            epoch_str = self._get_loss_and_acc_string(epoch=self.epochs_done, epochs=total_epochs,
                                                      train_loss=train_loss, train_acc=train_acc,
                                                      test_loss=test_loss, test_acc=test_acc,
                                                      batch=step % num_tr_iter, batches=num_tr_iter)
            self._manage_string(epoch_str, verbose, None)

    def _save_current_loss_and_acc(self, filename, train_loss, train_acc, test_loss, test_acc, step, fast_mode=False):
        """
        Adds the accuracy and loss result of current step to the pandas backup history and saves it (optional) into csv
        :param filename: File name where the data should be saved.
        :param train_loss:
        :param train_acc:
        :param test_loss:
        :param test_acc:
        :param step:
        :param fast_mode: Bool (Default False). If True, the file will not be saved.
        """
        a_series = pd.Series([step, self.epochs_done, train_loss, train_acc, test_loss, test_acc],
                             index=self.run_pandas.columns)
        self.run_pandas = self.run_pandas.append(a_series, ignore_index=True)
        if not fast_mode:
            if not filename.endswith('.csv'):
                filename += '.csv'
            filename = self.root_dir / filename
            self.run_pandas.to_csv(filename, index=False)

    def _tensorboard_checkpoint(self, x_train, y_train, x_test=None, y_test=None):
        """
        Saves the tensorboard data
        :param x_train: Train data
        :param y_train: Tran labels
        :param x_test: Test data (optional)
        :param y_test: Test labels (optional)
        :return: None
        """
        # Save train loss and accuracy
        train_loss, train_accuracy = self.evaluate(x_train, y_train)
        with self.train_summary_writer.as_default():
            tf.summary.scalar('loss', train_loss, step=self.epochs_done)
            tf.summary.scalar('accuracy', train_accuracy * 100, step=self.epochs_done)
        # Save test loss and accuracy
        if x_test is not None:
            assert y_test is not None
            test_loss, test_accuracy = self.evaluate(x_test, y_test)
            with self.test_summary_writer.as_default():
                tf.summary.scalar('loss', test_loss, step=self.epochs_done)
                tf.summary.scalar('accuracy', test_accuracy * 100, step=self.epochs_done)
        # Save weights histogram
        x = x_train
        for layer in self.shape:
            x = layer.save_tensorboard_checkpoint(x, self.weights_summary_writer,
                                                  self.activation_summary_writer, self.epochs_done)
        self._save_tensorboard_gradients(x_train, y_train)

    def _save_tensorboard_gradients(self, x_train, y_train):
        with tf.GradientTape() as tape:
            x_called = self.call(x_train)  # Forward mode computation
            current_loss = self._apply_loss(y_train, x_called)  # Compute loss
        variables = []
        for lay in self.shape:
            variables.extend(lay.trainable_variables())
        gradients = tape.gradient(current_loss, variables)  # Compute gradients
        assert all(g is not None for g in gradients)
        assert len(gradients) % 2 == 0, "No biases still not supported."  # TODO: what if you have no bias? This crashes
        with self.gradients_summary_writer.as_default():
            for i in range(int(len(gradients) / 2)):
                if gradients[2 * i].dtype == tf.complex64 or gradients[2 * i].dtype == tf.complex128:
                    tf.summary.histogram(name="Gradients_w_" + str(i) + "_real",
                                         data=tf.math.real(gradients[2 * i]), step=self.epochs_done)
                    tf.summary.histogram(name="Gradients_w_" + str(i) + "_imag",
                                         data=tf.math.imag(gradients[2 * i]), step=self.epochs_done)
                    tf.summary.histogram(name="Gradients_b_" + str(i) + "_real",
                                         data=tf.math.real(gradients[2 * i + 1]), step=self.epochs_done)
                    tf.summary.histogram(name="Gradients_b_" + str(i) + "_imag",
                                         data=tf.math.imag(gradients[2 * i + 1]), step=self.epochs_done)
                elif gradients[2 * i].dtype == tf.float32 or gradients[2 * i].dtype == tf.float64:
                    tf.summary.histogram(name="Gradients_w_" + str(i),
                                         data=gradients[2 * i], step=self.epochs_done)
                    tf.summary.histogram(name="Gradients_b_" + str(i),
                                         data=gradients[2 * i + 1], step=self.epochs_done)
                else:
                    logger.error("Input_dtype not supported. Should never have gotten here!", exc_info=True)
                    sys.exit(-1)

    def save(self, loss, acc):
        # https://stackoverflow.com/questions/2709800/how-to-pickle-yourself
        # TODO: TypeError: can't pickle _thread._local objects
        # https://github.com/tensorflow/tensorflow/issues/33283
        # loss, acc = self.evaluate(x, y)
        checkpoint_root = self.root_dir + "saved_models/"
        os.makedirs(checkpoint_root, exist_ok=True)
        save_name = checkpoint_root + "model_checkpoint_epoch" + str(self.epochs_done) + "loss{0:.4f}acc{1:d}".format(
            loss, int(acc * 100))

        # CvnnModel._extractfromlocal(self)
        with open(save_name.replace(".", ",") + ".pickle", "wb") as saver:
            # set_trace()
            pickle.dump(self.__dict__, saver)
            # for lay in self.shape:
            #    pickle.dump(lay.__dict__, saver)
        # CvnnModel._loadtolocal(self)

    # ================
    # Managing strings
    # ================

    def _manage_string(self, string, verbose="SILENT", filename=None, mode="a"):
        """
        Prints a string to console and/or saves it to a file
        :param string: String to be printed/saved
        :param verbose: String (Default 'SILENT'). If "DEBUG" it will print the message into the logger as info.
        :param filename: Filename where to save the string. If None it will not save it (default: None)
        :param mode: Mode to open the filename
        :return: None
        """
        if verbose in ("DEBUG",):
            logger.info(string)
        if filename is not None:
            filename = self.root_dir / filename
            try:
                with open(filename, mode) as file:
                    file.write(string + "\n")
            except FileExistsError:  # TODO: Check if this is the actual error
                logging.error("Same file already exists. Aborting to not override results" + str(filename),
                              exc_info=True)
            except FileNotFoundError:
                logging.error("No such file or directory: " + self.root_dir, exc_info=True)
                sys.exit(-1)

    def _get_str_evaluate(self, epoch, epochs, x, y, x_test=None, y_test=None, batch=None, batches=None) -> str:
        train_loss, train_acc, test_loss, test_acc = self.evaluate_train_and_test(x, y, x_test, y_test)
        return self._get_loss_and_acc_string(epoch, epochs, train_loss, train_acc, test_loss, test_acc, batch, batches)

    @staticmethod
    def _get_loss_and_acc_string(epoch, epochs, train_loss, train_acc, test_loss=None, test_acc=None, batch=None,
                                 batches=None) -> str:
        ret_str = "Epoch {0}/{1}".format(epoch, epochs)
        if batch is not None and batches is not None:
            ret_str += "; batch {0}/{1}".format(batch, batches)
        assert train_acc is not None and train_loss is not None, "Train loss or acc was None with fast_mode False"
        ret_str += " - train loss: {0:.4f} - train accuracy: {1:.2f} %".format(train_loss, train_acc * 100)
        if test_loss is not None and test_acc is not None:
            ret_str += "- test loss: {0:.4f} - test accuracy: {1:.2f} %".format(test_loss, test_acc * 100)
        return ret_str

    def summary(self):
        """
        Generates a string of a summary representation of your model.
        :return: string of the summary of the model
        """
        summary_str = ""
        summary_str += self.name + "\n"
        if self.is_complex():
            summary_str += "Complex Network\n"
        else:
            summary_str += "Real Network\n"
        for lay in self.shape:
            summary_str += lay.get_description()
        return summary_str

    def training_param_summary(self):
        summary_str = ""
        summary_str += self.name + "\n"
        if self.is_complex():
            summary_str += "Complex Network"
        else:
            summary_str += "Real Network"
        total_params = 0
        t = PrettyTable(['Layer (type)', 'Output Shape', 'Param #'])
        for lay in self.shape:
            train_params = 0
            for param in lay.trainable_variables():
                train_params += np.prod(param.shape)
            t.add_row([
                lay.__class__.__name__,
                lay.get_output_shape_description(),
                str(train_params)
            ])
            total_params += train_params
        print(summary_str)
        print(t)
        print("Total Trainable params: " + str(total_params))


if __name__ == '__main__':
    # monte_carlo_loss_gaussian_noise(iterations=100, filename="historgram_gaussian.csv")
    m = 5000
    n = 100
    coef_correls_list = np.linspace(-0.9, 0.9, 2)  # 4 classes
    param_list = []
    for coef in coef_correls_list:
        param_list.append([coef, 1, 2])
    dataset = dp.CorrelatedGaussianCoeffCorrel(m, n, param_list, debug=False)
    # x_fit = transform_to_real(dataset.x)      # To run real case
    x_fit = dataset.x  # To run complex case

    # Define shape
    cdtype = x_fit.dtype
    if cdtype == np.complex64:
        rdtype = np.float32
    else:
        rdtype = np.float64
    input_size = np.shape(x_fit)[1]
    hidden_size = 100
    output_size = np.shape(dataset.y_train)[1]
    # set_trace()
    shape = [layers.Dense(output_size=hidden_size, input_size=input_size, activation='cart_relu',
                          input_dtype=cdtype, dropout=None),
             layers.Dense(output_size=hidden_size, activation='cart_relu'),
             layers.Dense(output_size=hidden_size, activation='cart_relu'),
             layers.Dense(output_size=hidden_size, activation='cart_relu'),
             layers.Dense(output_size=output_size, activation='softmax_real')]
    # set_trace()
    # Train model
    model = CvnnModel("Testing_dropout", shape, tf.keras.losses.categorical_crossentropy,
                      tensorboard=False, verbose=False)
    model.fit(x_fit, dataset.y, validation_split=0.0, batch_size=100, epochs=10,
              verbose=True, save_csv_history=True)
    # start = time.time()
    # model.fit(dataset.x, dataset.y, batch_size=100, epochs=30, verbose=False)
    # end = time.time()
    # print(end - start)

    # Analyze data
    # model.plotter.plot_key(key='accuracy', showfig=False, savefig=True)
    # model.plotter.plot_key(key='loss', library='matplotlib', showfig=False, savefig=True)
    # model.get_confusion_matrix(dataset.x_test, dataset.y_test, save_result=True)
    # set_trace()

# How to comment script header
# https://medium.com/@rukavina.andrei/how-to-write-a-python-script-header-51d3cec13731
__author__ = 'J. Agustin BARRACHINA'
__copyright__ = 'Copyright 2020, {project_name}'
__credits__ = ['{credit_list}']
__license__ = '{license}'
__version__ = '0.2.40'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'
__status__ = '{dev_status}'
